/*******************************************************************************
**  FILE:           meshengine.h
**  DESCRIPTION:    This is the base class for computational engines which
**		    will take in a base Mesh object and provide methods for
**		    copying and allocating memory for the engine.
**
**  AUTHOR:         Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#ifndef MESHENGINE_H
#define MESHENGINE_H
#include <QtCore>
#include <QObject>
#include "QtOclMesh_global.h"
#include "point3d.h"
#include "size3d.h"


class Mesh;
class MeshItem;
class ProbeMeshItem;
class SourceMeshItem;



class QTOCLMESHSHARED_EXPORT MeshEngine : public QObject
{
    Q_OBJECT

public:
  enum MeshDirection{PX,NX,PY,NY,PZ,NZ};

protected:

  Mesh*                       mesh_;
  QList<ProbeMeshItem*>       meshProbes_;
  QList<SourceMeshItem*>      meshSources_;
  MeshSize                    meshSize_;
  QString                     engineName_;
  int                         dimensions_;
  int                         nodeVoltageCount_;
  int                         nodeBoundaryCount_;
  int                         nodePropertyCount_;

private:

  QReadWriteLock&             meshMutex_;
  QReadWriteLock              meshEngineMutex_;
  int                         currentIteration_;
  int                         totalIterations_;
  int                         updateInterval_;
  int                         delayFactor_;
  bool                        engineCompatible_;
  Point3D<float>              meshOrigin_;


public:
  MeshEngine( Mesh* initMesh );
  MeshEngine( const MeshEngine& other );
  ~MeshEngine();

  MeshEngine& operator=( const MeshEngine& other );

  QReadWriteLock& getEngineMutex();
  const QString getEngineName() const;
  const int dimensions() const;
  const MeshSize meshSize() const;
  const Size3D<float> meshResolution() const;
  const int getNodeVoltageCount() const;
  const int getNodeBoundaryCount() const;
  const int getNodePropertyCount() const;
  const int getCurrentIteration() const;

  void setCompatibility(const bool compatible);
  const bool isCompatible() const;


  void addMeshItem( MeshItem* meshItem );
  void addProbe( ProbeMeshItem* probe );
  void addSource( SourceMeshItem* source );
  const QList<ProbeMeshItem*>& getProbes() const;
  MeshPoint discretizePoint(const Point3D<float> point) const;


  void resetSimulationData();
  void deleteSimulationData();

  void clearVoltages();

protected:
  void startDataWrite();
  void endDataWrite();
  void startDataRead();
  void endDataRead();


private:
  void analyzeMeshItems(QList<MeshItem*>& meshItems);
  void setOuterMeshBoundary();


signals:
  void voltageDataChanged();
  void engineCompatibilityChanged( bool );



//-----------------------------------------------------------------------------
// VIRTUAL METHODS
//-----------------------------------------------------------------------------

public:
  virtual MeshEngine* clone() const = 0;

  virtual void setNodeBoundary( const MeshPoint& point,
                                const MeshDirection boundary,
                               const float value) = 0;

  virtual void setNodeConductance(const MeshPoint& point,
                                  const QVector<float> conductance) = 0;
  void setNodeConductance(const MeshPoint& point, const float conductance);

  virtual void setNodePermittity( const MeshPoint& point,
                               const QVector<float> permittivity) = 0;
  void setNodePermittity( const MeshPoint& point, const float permittivity);

  virtual void setNodePermeability( const MeshPoint& point,
                               const QVector<float> permeability) = 0;
  void setNodePermeability( const MeshPoint& point, const float permeability);

  virtual void setNodeVoltage( const MeshPoint& point,
                               const int voltage, const float value) = 0;
  virtual float& getNodeVoltage( const MeshPoint& point, const int voltage) = 0;

  virtual void createSimulation();
  virtual void simulateIteration( int iteration );
  virtual void updateHost();

  virtual void exciteMesh( int iteration ) = 0;
  virtual void sampleMesh( int iteration ) = 0;


protected:

  // Resource Management
  virtual void allocateResources( const MeshSize& meshSize ) = 0;
  virtual void initializeResources( const MeshSize& meshSize ) = 0;
  virtual void releaseResources() = 0;




};

#endif // MESHENGINE_H
