/*******************************************************************************
**  FILE:	    point3d.h
**  DESCRIPTION:    Represents a fundamental data structure used within throughout
**		    the qtoclmesh library.  It is a template point class which
**                  represents a generic position using the specified data type.
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#ifndef POINT3D_H
#define POINT3D_H
#include <QtCore>
#include "QtOclMesh_global.h"
#include <math.h>

template <typename T>
class /*QTOCLMESHSHARED_EXPORT*/ Point3D
{

private:

  T x, y, z;

public:

  Point3D() :
      x(0), y(0), z(0)
  {
  }



  Point3D(const T initX,const T initY) :
      x(initX), y(initY), z(0)
  {
  }



  Point3D(const T initX,const T initY,const T initZ) :
      x(initX), y(initY), z(initZ)
  {
  }



  Point3D(const Point3D<T>& other)
  {
    x = other.x;
    y = other.y;
    z = other.z;
  }



  //~Point3D()
  //{
  //}



  Point3D&
  operator=(const Point3D<T>& other)
  {
    if(&other!=this){
      x = other.x;
      y = other.y;
      z = other.z;
    }
    return *this;
  }



  bool
  operator>(const Point3D<T>& other) const
  {
    return (x > other.x)&&(y > other.y)&&(z > other.z);
  }



  bool
  operator>=(const Point3D<T>& other) const
  {
    return (x >= other.x)&&(y >= other.y)&&(z >= other.z);
  }



  bool
  operator<(const Point3D<T>& other) const
  {
    return (x < other.x)&&(y < other.y)&&(z < other.z);
  }



  bool
  operator<=(const Point3D<T>& other) const
  {
    return (x <= other.x)&&(y <= other.y)&&(z <= other.z);
  }



  bool
  operator==(const Point3D<T>& other) const
  {
    return (x == other.x)&&(y == other.y)&&(z == other.z);
  }



  bool
  operator!=(const Point3D<T>& other) const
  {
    return !(*this==other);
  }



  Point3D&
  operator+=(const Point3D<T>& other)
  {
    x += other.x;
    y += other.y;
    z += other.z;
    return *this;
  }



  Point3D&
  operator-=(const Point3D<T>& other)
  {
    x -= other.x;
    y -= other.y;
    z -= other.z;
    return *this;
  }



  Point3D&
  operator/=(const T value)
  {
    x /= value;
    y /= value;
    z /= value;
    return *this;
  }



  Point3D&
  operator*=(const T value)
  {
    x *= value;
    y *= value;
    z *= value;
    return *this;
  }



  Point3D&
  operator*(const T value) const
  {
    return Point3D<T>(*this) *= value;
  }



  Point3D&
  operator*=(const Point3D<T>& other)
  {
    x *= other.x;
    y *= other.y;
    z *= other.z;
    return *this;
  }



  Point3D&
  operator*(const Point3D<T>& other) const
  {
    return Point3D<T>(*this) *= other;
  }



  inline const T
  cx() const
  {
    return x;
  }



  inline const T
  cy() const
  {
    return y;
  }



  inline const T
  cz() const
  {
    return z;
  }



  inline T&
  rx()
  {
    return x;
  }



  inline T&
  ry()
  {
    return y;
  }



  inline T&
  rz()
  {
    return z;
  }



  float
  magnitude() const
  {
    return (float)sqrt(pow(x,2)+pow(y,2)+pow(z,2));
  }



  Point3D<T> mask() const
  {
    Point3D<T> maskPoint;
    maskPoint.rx() = (x == 0) ? 0 : 1;
    maskPoint.ry() = (y == 0) ? 0 : 1;
    maskPoint.rz() = (z == 0) ? 0 : 1;
    return maskPoint;
  }



  Point3D<float>
  unitVector() const
  {
    Point3D<float> unitPoint;
    // Need to force the conversion to float since
    // we won't know what type this object is storing
    unitPoint.rx() = (float)x;
    unitPoint.ry() = (float)y;
    unitPoint.rz() = (float)z;

    float mag = unitPoint.magnitude();
    unitPoint/=mag;
    return unitPoint;
  }



  Point3D<T>
  cross(const Point3D<T>& other) const
  {
    return Point3D<T>(y * other.z - z * other.y,
                      x * other.z - z * other.x,
                      x * other.y - y * other.x);
  }


};



template <typename T> Point3D<T>
operator+(const Point3D<T>& lhs, const Point3D<T>& rhs)
{
  Point3D<T> ret(lhs);
  ret.rx() += rhs.cx();
  ret.ry() += rhs.cy();
  ret.rz() += rhs.cz();
  return ret;
}



template <typename T> Point3D<T>
operator-(const Point3D<T>& lhs, const Point3D<T>& rhs)
{
  Point3D<T> ret(lhs);
  ret.rx() -= rhs.cx();
  ret.ry() -= rhs.cy();
  ret.rz() -= rhs.cz();
  return ret;
}



template <typename T> Point3D<T>
Min(const Point3D<T> &a, const Point3D<T> &b)
{
  Point3D<T> retPoint;
  retPoint.rx() = qMin(a.cx(),b.cx());
  retPoint.ry() = qMin(a.cy(),b.cy());
  retPoint.rz() = qMin(a.cz(),b.cz());
  return retPoint;
}



template <typename T> Point3D<T>
Max(const Point3D<T> &a, const Point3D<T> &b)
{
  Point3D<T> retPoint;
  retPoint.rx() = qMax(a.cx(),b.cx());
  retPoint.ry() = qMax(a.cy(),b.cy());
  retPoint.rz() = qMax(a.cz(),b.cz());
  return retPoint;
}



template <typename T> T
dot(const Point3D<T> &a, const Point3D<T> &b)
{
  return a.cx() * b.cx() + a.cy() * b.cy() + a.cz() * b.cz();
}



template <typename T> Point3D<T>
cross(const Point3D<T> &a, const Point3D<T> &b)
{
  return a.cross(b);
}



template <typename T> QDebug
operator<<(QDebug dbg,const Point3D<T>& point)
{
  dbg.nospace()<<"Point(x,y,z):("<<point.cx()<<","<<point.cy()<<","<<point.cz()<<")";
  return dbg.space();
}



template <typename T> QDataStream&
operator<<(QDataStream &out, const Point3D<T>& point)
{
  out << point.cx()
      << point.cy()
      << point.cz();
  return out;
}



template <typename T> QDataStream&
operator>>(QDataStream &in, Point3D<T>& point)
{
  T tmp[3];
  in  >> tmp[0]
      >> tmp[1]
      >> tmp[2];

  point.rx() = tmp[0];
  point.ry() = tmp[1];
  point.rz() = tmp[2];

  return in;
}



typedef Point3D<int> MeshPoint;



#endif // POINT3D_H
