

/*********************************************************
**
**  Scattering Kernel
**
*********************************************************/
__kernel void scatter(__global float* mesh
                      , __global float* properties
                      , const unsigned int width
                      , const unsigned int height
                      , const unsigned int depth
                      , __local float* subMesh
)
{

  // The following code was an attempt at a shared memory implementation
  // which would be much faster however There are some issues with setting the
  // shared memory arguments size.
  bool valid = (get_global_id(0)<width)&(get_global_id(1)<height);
  int subMeshIndex;
  float y0;
  float y;
  float a = 0;
  if(valid){
    y0 = properties[get_global_id(1)*width+get_global_id(0)];
    y = y0 + properties[width*height+get_global_id(1)*width+get_global_id(0)] + 4;
    // Add voltages to local submesh.
    // Also calculate the 'a' variable which allows each scattered voltage to be
    // calculated using the least number of FLOP's
    for(int i=0;i<5;++i){
      subMeshIndex = i*get_local_size(0)*get_local_size(1)+get_local_id(1)*get_local_size(0)+get_local_id(0);
      subMesh[subMeshIndex] = mesh[i*width*height+get_global_id(1)*width+get_global_id(0)];
      if(i==4){
        a += subMesh[subMeshIndex]*y0;
      }else{
        a += subMesh[subMeshIndex];
      }
    }
    a = a*2.0/y;
  }
  write_mem_fence(CLK_LOCAL_MEM_FENCE);

  if(valid){
    // Scatter voltages in submesh
    for(int i=0;i<5;++i){
      subMeshIndex = i*get_local_size(0)*get_local_size(1)+get_local_id(1)*get_local_size(0)+get_local_id(0);
      mesh[i*width*height+get_global_id(1)*width+get_global_id(0)] = a - subMesh[subMeshIndex];
    }
  }

  // Global memory thread synchronization (i.e. sync global mesh writes)
//  barrier(CLK_GLOBAL_MEM_FENCE);

}




/*********************************************************
**
**  Impulse Interchange Kernel
**
*********************************************************/

__kernel void impulseInterchange(
    __global float* mesh,                       // Mesh Voltages
    __global unsigned short* boundaries,      // Mesh Boundaries
    __global float* boundaryTable,            // constant boundary lookup table
    const unsigned int width,
    const unsigned int height,
    const unsigned int depth,
    __local float* subMesh
    )
{
  int gIDX = get_global_id(0);
  int gIDY = get_global_id(1);
  bool valid = (gIDX < width)&(gIDY < height);
  float tmpVoltage;
  unsigned short boundaryCode;
  int subMeshIndex;
  int otherVoltage;
  int otherSubMeshIndex;

  // load voltages into the local submesh
  for(int i = 0; i < 4; ++i){
    subMeshIndex = i*get_local_size(0)*get_local_size(1) + get_local_id(1)*get_local_size(0) + get_local_id(0);
    subMesh[subMeshIndex] = valid*mesh[i*width*height + gIDY*width + gIDX];
  }

  write_mem_fence(CLK_LOCAL_MEM_FENCE);

  if(valid){
    // Loop through and interchange impulses - this is implemented by only
    // interchanging in two directions: interchange below, and to the right
    for(int i = 0; i < 4; ++i){
      boundaryCode = boundaries[i*width*height + gIDY*width + gIDX];
      subMeshIndex = i*get_local_size(0)*get_local_size(1) + get_local_id(1)*get_local_size(0) + get_local_id(0);
      tmpVoltage = 0;
      // If no boundary exists at this node, then exchange with node below or the right
      // Exchange impulses with node - Do this by first checking if this
      // work item is on the work group boundary, if it is then the impulse has
      // to be exchanged with the one in global memory from the neighbouring
      // work group.  Otherwise use the local subMesh.
      if(boundaryCode == 0){
        if(i == 0){
          otherVoltage = 2;
          tmpVoltage = subMesh[subMeshIndex];
          if(get_local_id(1) < get_local_size(1)-1){
            otherSubMeshIndex = otherVoltage*get_local_size(0)*get_local_size(1) + (get_local_id(1)+1)*get_local_size(0) + get_local_id(0);
            subMesh[subMeshIndex] = subMesh[otherSubMeshIndex];
            subMesh[otherSubMeshIndex] = tmpVoltage;
          }
        }
        else if(i == 3){
          otherVoltage = 1;
          tmpVoltage = subMesh[subMeshIndex];
          if(get_local_id(0) < get_local_size(0)-1){
            otherSubMeshIndex = otherVoltage*get_local_size(0)*get_local_size(1) + get_local_id(1)*get_local_size(0) + get_local_id(0) + 1;
            subMesh[subMeshIndex] = subMesh[otherSubMeshIndex];
            subMesh[otherSubMeshIndex] = tmpVoltage;
          }

        }
      }
      else{
        subMesh[subMeshIndex] *= boundaryTable[boundaryCode];
      }
    }
  }
  //barrier(CLK_GLOBAL_MEM_FENCE);
  write_mem_fence(CLK_LOCAL_MEM_FENCE);


  // Put all of the voltages back into global memory
  if(valid){
    for(int i = 0; i < 4; ++i){
      subMeshIndex = i*get_local_size(0)*get_local_size(1) + get_local_id(1)*get_local_size(0) + get_local_id(0);
      mesh[i*width*height + gIDY*width + gIDX] = subMesh[subMeshIndex];
    }
  }

//  barrier(CLK_GLOBAL_MEM_FENCE);
  write_mem_fence(CLK_GLOBAL_MEM_FENCE);

  if(valid){
    if(get_local_id(1) == get_local_size(1) - 1){
      // memory access is based on: (voltage Arm)*(mesh area) + globalIDY*width + globalIDX
      // here voltage arm = 0
      boundaryCode = boundaries[gIDY*width + gIDX];
      if(boundaryCode == 0){
        tmpVoltage = mesh[gIDY*width + gIDX];
        mesh[gIDY*width + gIDX] = mesh[2*width*height + (gIDY+1)*width + gIDX];
        mesh[2*width*height + (gIDY+1)*width + gIDX] = tmpVoltage;
      }

    }
    if(get_local_id(0) == get_local_size(0) - 1){
      boundaryCode = boundaries[3*width*height + gIDY*width + gIDX];
      if(boundaryCode == 0){
        tmpVoltage = mesh[3*width*height + gIDY*width + gIDX];
        mesh[3*width*height + gIDY*width + gIDX] = mesh[1*width*height + gIDY*width + gIDX + 1];
        mesh[1*width*height + gIDY*width + gIDX + 1] = tmpVoltage;
      }
    }
  }
   barrier(CLK_GLOBAL_MEM_FENCE);
}





/*********************************************************
**
**  Sampling Kernel
**
*********************************************************/


__kernel void sample(
    __global float* mesh              // Mesh Voltages
  , __global uint* probeInfo          // Coordinates for each probe
  , __global float* probeData         // Container for the samples taken
  , unsigned int probeInfoStride      // The stride of the probe information
  , unsigned int probeBufferSize      // Buffer Size
  , __global float* properties        // Node Properties
  , unsigned int iteration            // Current iteration
  , unsigned int meshWidth
  , unsigned int meshHeight
  , unsigned int meshDepth
    )
{
  int x = probeInfo[get_global_id(0)*probeInfoStride];
  int y = probeInfo[get_global_id(0)*probeInfoStride+1];
  float avgVoltage = 0;
  float y0 = properties[y*meshWidth + x];
  unsigned int meshArea = meshWidth*meshHeight;
  for(int i = 0; i < 5; ++i){
    if(i < 4){
      avgVoltage += mesh[i*meshArea + y*meshWidth + x];
    }else{
      avgVoltage += y0*mesh[i*meshArea + y*meshWidth + x];
    }
  }
  avgVoltage *= 2.0/(y0 + 4.0);
  probeData[get_global_id(0)*probeBufferSize + iteration % probeBufferSize] = avgVoltage;

  barrier(CLK_GLOBAL_MEM_FENCE);
}




/*********************************************************
**
**  Excitation Kernel
**
*********************************************************/


/******************************************
** sourceInfo:
** [0][1][2]	=> [x][y][z]
** [3][4][5]	=> [width][height][depth]
** [6]		=> Duration
** [7]		=> Direction code
** [8]		=> Dataoffset
*******************************************/


__kernel void excite(
    __global float* mesh,                      // Mesh Voltages
    __global unsigned int* sourceInfo,         // Source Coordinates and sizes
    __global float* sourceData,                // Source Data
    unsigned int numSources,                    // Number of sources
    int sourceInfoStride,	    		// Stride of source info vector
    unsigned int iteration,
    unsigned int meshWidth,
    unsigned int meshHeight,
    unsigned int meshDepth
    )
{

  int sourceNum = get_local_id(0);

  if(sourceInfo[sourceInfoStride*sourceNum+6]>=iteration){
    unsigned int x = sourceInfo[sourceInfoStride*sourceNum];
    unsigned int y = sourceInfo[sourceInfoStride*sourceNum+1];
    //    unsigned int sourceWidth = sourceInfo[sourceInfoStride*sourceNum+3];
    //    unsigned int sourceHeight = sourceInfo[sourceInfoStride*sourceNum+4];
    float value = sourceData[sourceInfo[sourceInfoStride*sourceNum+8]+iteration-1];
    //    unsigned int direction = sourceInfo[sourceInfoStride*sourceNum+7];

    for(int i = 0;i<4;++i){
      //mesh[i*meshWidth*meshHeight+y*meshWidth+x] += ((direction>>(i))&0x1)*value;
      mesh[i*meshWidth*meshHeight+y*meshWidth+x] += value;
    }
  }

//barrier(CLK_GLOBAL_MEM_FENCE);
}

