/*******************************************************************************
**  FILE:
**  DESCRIPTION:
**
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include "inc/boundarymeshitems.h"
#include "inc/mesh.h"



/**********************************************************
**
**  BOUNDARY MESH ITEM
**
**********************************************************/



//-----------------------------------------------------------------------------
// CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------


BoundaryMeshItem::BoundaryMeshItem():
	MeshItem()
{
  displayText_    = QString("Boundary");
  whatsThisText_  = QString("Boundary mesh items for "
                           "representing electromagnetic boundaries.");
  toolTipText_    = QString("Boundary Items");
  fillColor_      = QColor(Qt::red);
  borderColor_    = QColor(Qt::black);

  addProperty(QString("reflection coefficient"),QString("R"),QVariant(-1.0f));
}



BoundaryMeshItem::BoundaryMeshItem(const BoundaryMeshItem &other):
	MeshItem(other)
{
}



BoundaryMeshItem::~BoundaryMeshItem()
{
}



//-----------------------------------------------------------------------------
// OPERATOR OVERLOADS
//-----------------------------------------------------------------------------


BoundaryMeshItem&
BoundaryMeshItem::operator=(const BoundaryMeshItem& other)
{
  if(&other!=this){
    MeshItem::operator =(other);
  }
  return *this;
}



MeshItem*
BoundaryMeshItem::clone() const
{
  return new BoundaryMeshItem(*this);
}



void
BoundaryMeshItem::setMesh(Mesh *mesh)
{
  MeshItem::setMesh(mesh);
  if(this->mesh() != 0){
    Size3D<float> res = this->mesh()->getMeshResolution()/3.0;
    setMinimumSize(res);
  }
}


BoundaryMeshItem::operator QVariant() const
{
  return QVariant::fromValue(*this);
}



// Wrapper around the meshEngine so that we can use this method instead
// of operating directly to the meshEngine.
void
BoundaryMeshItem::setNodeBoundary(MeshEngine& meshEngine,
                                  MeshPoint pos,
                                  MeshEngine::MeshDirection direction)
{
  float reflectionCoef =
      getPropertyValue(QString("BoundaryMeshItem/reflection coefficient")).toFloat();
  meshEngine.setNodeBoundary(pos,direction,reflectionCoef);

}


//-----------------------------------------------------------------------------
// STREAM OPERATORS
//-----------------------------------------------------------------------------


QDebug
operator<<(QDebug dbg,const BoundaryMeshItem& item){
  dbg.nospace()<<"\nBoundary Mesh Item:"<< (MeshItem)item;
  return dbg.maybeSpace();
}


QDataStream&
operator<<(QDataStream &out, const BoundaryMeshItem& item){
  out << (MeshItem)item;
  return out;
}


QDataStream&
operator>>(QDataStream &in, BoundaryMeshItem& item){
  MeshItem tmpItem;
  in>>tmpItem;
  (MeshItem)item = tmpItem;
  return in;
}





/**********************************************************
**
**  LINEAR BOUNDARY MESH ITEM
**
**********************************************************/


//-----------------------------------------------------------------------------
//  CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------


LinearBoundary::LinearBoundary()
  : BoundaryMeshItem()
{
  displayText_    = QString("Line");
  whatsThisText_  = QString("Mesh Item representing a linear "
                           "boundary in a dielectric mesh.");
  toolTipText_    = QString("Linear Dielectric Boundary Item");
  fillColor_      = QColor(200,0,0,255);
}



LinearBoundary::LinearBoundary(const LinearBoundary &other)
  : BoundaryMeshItem(other)
{
}



LinearBoundary::~LinearBoundary()
{
}



//-----------------------------------------------------------------------------
//  OPERATOR OVERLOADS
//-----------------------------------------------------------------------------


LinearBoundary&
LinearBoundary::operator =(const LinearBoundary &other)
{
  if(&other!=this){
    BoundaryMeshItem::operator =(other);
  }
  return *this;
}



MeshItem*
LinearBoundary::clone() const
{
  return new LinearBoundary(*this);
}



LinearBoundary::operator QVariant() const
{
  return QVariant::fromValue(*this);
}



//-----------------------------------------------------------------------------
//  VIRTUAL METHOD DEFINITIONS
//-----------------------------------------------------------------------------


void
LinearBoundary::setSize(const Size3D<float> size)
{
  Size3D<GLfloat> res = minimumSize();
  Size3D<float> newSize = size;

//  if(up().mask().cx() == 1){
//    newSize.setHeight((qAbs(size.height()) >= qAbs(size.depth()) ) ? size.height() : res.height()) ;
//    newSize.setDepth((qAbs(size.depth()) > qAbs(size.height()) ) ? size.depth() : res.depth());
//  }
//  else if(up().mask().cy() == 1){
//    newSize.setWidth((qAbs(size.width()) >= qAbs(size.depth()) ) ? size.width() : res.width());
//    newSize.setDepth((qAbs(size.depth()) > qAbs(size.width()) ) ? size.depth() : res.depth());
//  }
//  else if(up().mask().cz() == 1){
//    newSize.setWidth((qAbs(size.width()) >= qAbs(size.height()) ) ? size.width() : res.width());
//    newSize.setHeight((qAbs(size.height()) > qAbs(size.width()) ) ? size.height() : res.height());
//  }

  MeshItem::setSize(newSize);
}



void
LinearBoundary::addToMesh(MeshEngine &meshEngine)
{
  MeshPoint bottomBackLeft = meshEngine.discretizePoint(getBottomBackLeft(this));
  MeshPoint topFrontRight = meshEngine.discretizePoint(getTopFrontRight(this));

  // The discretizePoint function takes into account the fact that the Y axis
  // should be flipped to make the mesh appear the same as in the editor
  // this should probably be done with transforms instead of this method.
  int tmpVal = bottomBackLeft.cy();
  bottomBackLeft.ry() = topFrontRight.cy();
  topFrontRight.ry() = tmpVal;

//  qDebug()<<"BBL:"<<bottomBackLeft<<"TFR:"<<topFrontRight;
  MeshPoint point = bottomBackLeft;
  MeshPoint pointDifference = topFrontRight - bottomBackLeft;
  if(pointDifference.rx() != 0){
    for(point.rx(); point.rx() < topFrontRight.cx(); point.rx()++){
      setNodeBoundary(meshEngine,point,MeshEngine::PY);
    }
  }
  else if(pointDifference.ry() != 0){
    for(point.ry(); point.ry() < topFrontRight.cy(); point.ry()++){
      setNodeBoundary(meshEngine,point,MeshEngine::NX);
    }
  }
//  else if(pointDifference.rz() != 0){
//    for(point.ry() = bottomBackLeft.cy(); point.ry() <= topFrontRight.cy(); point.ry()++){
//      qDebug()<<"Point:"<<point<<"Boundary:"<<MeshEngine::PY;
//    }
//  }
}



void
LinearBoundary::draw()
{
  Point3D<GLfloat> bottomBackLeft = this->pos();//getBottomBackLeft(this);
  Point3D<GLfloat>* glPoint;
  Point3D<GLfloat>* startPointPtr;
  Point3D<GLfloat>* endPointPtr;
  int facePointStride = 4;

//  GLfloat shapeHeight = qAbs(size().height());
//  GLfloat shapeWidth = qAbs(size().width());
//  GLfloat shapeDepth = qAbs(size().depth());

  Size3D<GLfloat> res = mesh()->getMeshResolution()/3.0;
  Size3D<GLfloat> size = this->size();

  Point3D<GLfloat> shiftAmmount;

//  GLfloat shapeHeight = (qAbs(size.height()) > res.height())
//      ? qAbs(size.height()) : res.height();
//  GLfloat shapeWidth = (qAbs(size.width()) > res.width())
//      ? qAbs(size.width()) : res.width();
//  GLfloat shapeDepth = (qAbs(size.depth()) > res.depth())
//      ? qAbs(size.depth()) : res.depth();

  GLfloat shapeWidth = size.width();
  GLfloat shapeHeight = size.height();
  GLfloat shapeDepth = size.depth();

  if(up().mask().cx() == 1){
    shapeHeight =  (qAbs(size.height()) >= qAbs(size.depth()) ) ? size.height() : res.height() ;
    shapeDepth = (qAbs(size.depth()) > qAbs(size.height()) ) ? size.depth() : res.depth() ;
  }
  else if(up().mask().cy() == 1){
    shapeWidth = (qAbs(size.width()) >= qAbs(size.depth()) ) ? size.width() : res.width() ;
    shapeDepth = (qAbs(size.depth()) > qAbs(size.width()) ) ? size.depth() : res.depth() ;
  }
  else if(up().mask().cz() == 1){
    shapeWidth = (qAbs(size.width()) >= qAbs(size.height()) ) ? size.width() : res.width() ;
    shapeHeight = (qAbs(size.height()) > qAbs(size.width()) ) ? size.height() : res.height() ;
  }

  if(shapeWidth == res.width()){
    shiftAmmount.rx() = res.width()/2;
  }
  if(shapeHeight == res.height()){
    shiftAmmount.ry() = res.height()/2;
  }
  if(shapeDepth == res.depth()){
    shiftAmmount.rz() = res.depth()/2;
  }

  bottomBackLeft -= shiftAmmount;

  QVector< Point3D<GLfloat> > facePoints;
  QVector< Point3D<GLfloat> > faceNormals;
  QVector< Point3D<GLfloat> > faceOffsets;

  // BOTTOM FACE
  facePoints << bottomBackLeft;
  facePoints << bottomBackLeft + Point3D<GLfloat>(shapeWidth,0,0);
  facePoints << bottomBackLeft + Point3D<GLfloat>(shapeWidth,0,shapeDepth);
  facePoints << bottomBackLeft + Point3D<GLfloat>(0,0,shapeDepth);
  // Normals for top and bottom faces
  faceNormals << Point3D<GLfloat>(0,-1.0,0) << Point3D<GLfloat>(0,1.0,0);
  // the bottom face offset
  faceOffsets << Point3D<GLfloat>(0,shapeHeight,0);

  // LEFT FACE
  facePoints << bottomBackLeft;
  facePoints << bottomBackLeft + Point3D<GLfloat>(0,0,shapeDepth);
  facePoints << bottomBackLeft + Point3D<GLfloat>(0,shapeHeight,shapeDepth);
  facePoints << bottomBackLeft + Point3D<GLfloat>(0,shapeHeight,0);
  // Normals for left and right faces
  faceNormals << Point3D<GLfloat>(-1.0,0,0) << Point3D<GLfloat>(1.0,0,0);
  // the right face offset
  faceOffsets << Point3D<GLfloat>(shapeWidth,0,0);


  // BACK FACE
  facePoints << bottomBackLeft;
  facePoints << bottomBackLeft + Point3D<GLfloat>(0,shapeHeight,0);
  facePoints << bottomBackLeft + Point3D<GLfloat>(shapeWidth,shapeHeight,0);
  facePoints << bottomBackLeft + Point3D<GLfloat>(shapeWidth,0,0);
  // Normals for front and back faces
  faceNormals << Point3D<GLfloat>(0,0,-1.0) << Point3D<GLfloat>(0,0,1.0);
  // the back face offset
  faceOffsets << Point3D<GLfloat>(0,0,shapeDepth);


  Point3D<float>* normalPtr = faceNormals.data();
  Point3D<float>* offsetPtr = faceOffsets.data();

  QColor fill = getFillColor();


  glBegin(GL_QUADS);
  glColor3f(fill.redF(),fill.greenF(),fill.blueF());
  for(startPointPtr = facePoints.data();
  startPointPtr <= facePoints.data() + (facePoints.count()-facePointStride);
  startPointPtr += facePointStride){
    // Set the ending point in the point vector for these two faces
    endPointPtr = startPointPtr + facePointStride - 1;

    glNormal3f(normalPtr->cx(), normalPtr->cy(), normalPtr->cz());
    normalPtr++;
    for(glPoint = startPointPtr; glPoint <= endPointPtr; glPoint++){
      glVertex3f(glPoint->cx(),glPoint->cy(),glPoint->cz());
    }

    glNormal3f(normalPtr->cx(), normalPtr->cy(), normalPtr->cz());
    normalPtr++;
    for(glPoint = endPointPtr; glPoint >= startPointPtr; glPoint--){
      glVertex3f(glPoint->cx() + offsetPtr->cx(),
                 glPoint->cy() + offsetPtr->cy(),
                 glPoint->cz() + offsetPtr->cz());
    }
    offsetPtr++;
  }
  glEnd();
}



const QPixmap
LinearBoundary::generateImage(const QSize &size,
                                        const QColor border,
                                        const QColor fill)
{
  QPixmap pixmap = getPixmap(size);
  QPainter painter(&pixmap);
  setupPainter(painter,border,fill);

  QRectF displayRect(QPointF(0,0),size);
  painter.drawLine(displayRect.topLeft(),displayRect.bottomRight());

  return pixmap;
}



//-----------------------------------------------------------------------------
//  STREAM OPERATORS
//-----------------------------------------------------------------------------


QDataStream&
operator<<(QDataStream &out, const LinearBoundary& item)
{
    out << (BoundaryMeshItem)item;
    return out;
}



QDataStream&
operator>>(QDataStream &in, LinearBoundary& item)
{
    BoundaryMeshItem tmpItem;
    in>>tmpItem;
    (BoundaryMeshItem)item = tmpItem;
    return in;
}



QDebug
operator<<(QDebug dbg,const LinearBoundary& item)
{
    dbg.nospace()<<"Linear Boundary Mesh Item:"<< (BoundaryMeshItem)item;
    return dbg.maybeSpace();
}





/**********************************************************
**
**  RECTANGLE DIELECTRIC BOUNDARY MESH ITEM
**
**********************************************************/


//-----------------------------------------------------------------------------
//  CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------


RectangleBoundary::RectangleBoundary():
    BoundaryMeshItem()
{
      displayText_    = QString("Rectangle");
      whatsThisText_  = QString("Mesh Item representing a rectangular "
                                        "boundary in a dielectric mesh.");
      toolTipText_    = QString("Rectangular Dielectric Boundary Item");
      fillColor_      = QColor(0,0,0,0);
}



RectangleBoundary::RectangleBoundary(
    const RectangleBoundary &other):
	BoundaryMeshItem(other)
{
  for(int i = 0; i < 4; ++i){
    walls[i] = other.walls[i];
  }
}



RectangleBoundary::~RectangleBoundary()
{
}


//-----------------------------------------------------------------------------
//  OPERATOR OVERLOADS
//-----------------------------------------------------------------------------


RectangleBoundary&
RectangleBoundary::operator=(const RectangleBoundary &other)
{
  if(&other!=this){
    BoundaryMeshItem::operator =(other);
    for(int i = 0; i < 4; ++i){
      walls[i] = other.walls[i];
    }
  }
  return *this;
}



MeshItem*
RectangleBoundary::clone() const
{
  return new RectangleBoundary(*this);
}



RectangleBoundary::operator QVariant() const
{
  return QVariant::fromValue(*this);
}



//-----------------------------------------------------------------------------
//  VIRTUAL METHOD DEFINITIONS
//-----------------------------------------------------------------------------


void
RectangleBoundary::setPos(const Point3D<float> pos)
{
  Point3D<float> sizePoint = size().toPoint3D();
  walls[0].setPos(pos);
  walls[1].setPos(pos);
  walls[2].setPos(pos+sizePoint*Point3D<float>(0,1,0));
  walls[3].setPos(pos+sizePoint*Point3D<float>(1,0,0));
  MeshItem::setPos(pos);
}



void
RectangleBoundary::setUp(const Point3D<float> up)
{
//  Point3D<float> topUp, sideUp;
////  if(up.cx() == 1){
////    topUp = Point3D<float>(0,1,0);
////    sideUp = Point3D<float>(0,0,1);
////  }
////  else if(up.cy() == 1){
////    topUp = Point3D<float>(0,0,1);
////    sideUp = Point3D<float>(1,0,0);
////  }
////  else if(up.cz() == 1){
////    topUp = Point3D<float>(0,1,0);
////    sideUp = Point3D<float>(1,0,0);
////  }

//  walls[0].setUp(topUp);
//  walls[1].setUp(sideUp);
//  walls[2].setUp(topUp);
//  walls[3].setUp(sideUp);

  walls[0].setUp(up);
  walls[1].setUp(up);
  walls[2].setUp(up);
  walls[3].setUp(up);

  MeshItem::setUp(up);

}



void
RectangleBoundary::setSize(const Size3D<float> size)
{
  setPos(pos());
  Size3D<float> wallSize = size;
  wallSize.setHeight(0);
  walls[0].setSize(wallSize);
  walls[2].setSize(wallSize);
  wallSize = size;
  wallSize.setWidth(0);
  walls[1].setSize(wallSize);
  walls[3].setSize(wallSize);

  MeshItem::setSize(size);
}



void
RectangleBoundary::addToMesh(MeshEngine &meshEngine)
{
  QVariant reflectionCoef = getPropertyValue(QString("BoundaryMeshItem/reflection coefficient"));
  for(int i = 0; i <= 3; ++i){
    walls[i].setPropertyValue(QString("BoundaryMeshItem/reflection coefficient"),reflectionCoef);
    walls[i].addToMesh(meshEngine);
  }
}


void
RectangleBoundary::setMesh(Mesh *mesh)
{
  for(int i = 0; i <= 3; ++i){
    walls[i].setMesh(mesh);
  }
  BoundaryMeshItem::setMesh(mesh);
}


void
RectangleBoundary::draw()
{
  for(int i = 0; i <= 3; ++i){
    walls[i].setSelected(isSelected());
    walls[i].draw();
  }
}



const QPixmap
RectangleBoundary::generateImage(const QSize &size,
                                           const QColor border,
                                           const QColor fill)
{
  QPixmap pixmap = getPixmap(size);
  QPainter painter(&pixmap);
  setupPainter(painter,border,fill);

  QRectF displayRect(QPointF(0,0),size);
  painter.drawRect(displayRect);
  return pixmap;
}



//-----------------------------------------------------------------------------
//  STREAM OPERATORS
//-----------------------------------------------------------------------------


QDataStream&
operator<<(QDataStream &out, const RectangleBoundary& item)
{
  out << (BoundaryMeshItem)item;
  return out;
}


QDataStream&
operator>>(QDataStream &in, RectangleBoundary& item)
{
  BoundaryMeshItem tmpItem;
  in >> tmpItem;
  (BoundaryMeshItem)item = tmpItem;
  return in;
}


QDebug
operator<<(QDebug dbg,const RectangleBoundary& item)
{
  dbg.nospace()<<"\nRectangular Dielectric Boundary Mesh Item:"<< (BoundaryMeshItem)item;
  return dbg.maybeSpace();
}



