/*******************************************************************************
**  FILE:           meshengine.cpp
**  DESCRIPTION:    This is the base class for computational engines which
**		    will take in a base Mesh object and provide methods for
**		    copying and allocating memory for the engine.
**
**  AUTHOR:         Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include <QtGui>
#include "inc/meshengine.h"
#include "inc/mesh.h"
#include "inc/meshitems.h"


//-----------------------------------------------------------------------------
// CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------


MeshEngine::MeshEngine(Mesh *mesh)
  : mesh_               ( mesh )
  , meshMutex_          ( mesh_->getMeshMutex() )
  , currentIteration_   ( 0 )
  , totalIterations_    ( 0 )
  , updateInterval_     ( 0 )
  , delayFactor_        ( 0 )
  , engineCompatible_   ( false )
  , engineName_         ( tr("Mesh Engine Base Class"))
  , nodeVoltageCount_   (0)
  , nodeBoundaryCount_  (0)
  , nodePropertyCount_  (0)
{

}



MeshEngine::MeshEngine(const MeshEngine &other)
  : mesh_             ( other.mesh_ )
  , meshProbes_       ( other.meshProbes_ )
  , meshSources_      ( other.meshSources_ )
  , meshSize_         ( other.meshSize_ )
  , meshMutex_        ( other.meshMutex_ )
  , currentIteration_	( other.currentIteration_ )
  , totalIterations_	( other.totalIterations_ )
  , updateInterval_   ( other.updateInterval_ )
  , delayFactor_      ( other.delayFactor_ )
  , engineCompatible_	( other.engineCompatible_ )
  , engineName_       ( other.engineName_ )
  , nodeVoltageCount_ ( other.nodeVoltageCount_ )
  , nodeBoundaryCount_( other.nodeBoundaryCount_ )
  , nodePropertyCount_( other.nodePropertyCount_ )
{
}



MeshEngine::~MeshEngine()
{

}



//-----------------------------------------------------------------------------
// PUBLIC METHODS
//-----------------------------------------------------------------------------



MeshEngine&
MeshEngine::operator =( const MeshEngine& other )
{
  if(this!=&other){
    startDataWrite();
    mesh_               = other.mesh_;
    currentIteration_   = other.currentIteration_;
    totalIterations_    = other.totalIterations_;
    updateInterval_	    = other.updateInterval_;
    delayFactor_        = other.delayFactor_;
    engineCompatible_   = other.engineCompatible_;
    meshSize_           = other.meshSize_;
    nodeVoltageCount_   = other.nodeVoltageCount_;
    nodeBoundaryCount_  = other.nodeBoundaryCount_;
    nodePropertyCount_  = other.nodePropertyCount_;
    releaseResources();
    endDataWrite();
    emit engineCompatibilityChanged(engineCompatible_);
  }
  return *this;
}



QReadWriteLock&
MeshEngine::getEngineMutex()
{
  return meshEngineMutex_;
}



const QString
MeshEngine::getEngineName() const
{
  return engineName_;
}



const int
MeshEngine::dimensions() const
{
  return dimensions_;
}



const MeshSize
MeshEngine::meshSize() const
{
  return meshSize_;
}



const Size3D<float>
MeshEngine::meshResolution() const
{
  return mesh_->getMeshResolution();
}



const int
MeshEngine::getNodeVoltageCount() const
{
  return nodeVoltageCount_;
}



const int
MeshEngine::getNodeBoundaryCount() const
{
  return nodeBoundaryCount_;
}



const int
MeshEngine::getNodePropertyCount() const
{
  return nodePropertyCount_;
}



const int
MeshEngine::getCurrentIteration() const
{
  return currentIteration_;
}


void
MeshEngine::setCompatibility( const bool compatible )
{
  engineCompatible_ = compatible;
  emit engineCompatibilityChanged(engineCompatible_);
}



const bool
MeshEngine::isCompatible() const
{
  return engineCompatible_;
}



void
MeshEngine::addMeshItem( MeshItem *meshItem )
{
  meshItem->addToMesh(*this);
}



void
MeshEngine::addProbe( ProbeMeshItem *probe )
{
  if( meshProbes_.contains(probe) == false){
    meshProbes_.append(probe);
  }
  else{
    qDebug()<<"Probe has already been added to Mesh Engine.";
  }
}



void
MeshEngine::addSource( SourceMeshItem *source )
{
  if( meshSources_.contains(source) == false){
    meshSources_.append(source);
  }
  else{
    qDebug()<<"Source has already been added to Mesh Engine.";
  }
}


const QList<ProbeMeshItem*>&
MeshEngine::getProbes() const
{
  return meshProbes_;
}



MeshPoint
MeshEngine::discretizePoint(const Point3D<float> point) const
{
  Point3D<float> newPoint = point - meshOrigin_;
  MeshPoint retPoint;
  retPoint.rx() = qRound((newPoint.cx()/mesh_->getMeshResolution().width()));
  retPoint.ry() = qRound((newPoint.cy()/mesh_->getMeshResolution().height()));
  retPoint.rz() = qRound((newPoint.cz()/mesh_->getMeshResolution().depth()));

  retPoint.ry() = qAbs(retPoint.ry()-(meshSize_.height()));
  return retPoint;
}



void
MeshEngine::resetSimulationData()
{
  startDataWrite();
  foreach(ProbeMeshItem* probe, meshProbes_){
    probe->clearData();
  }
  currentIteration_ = 0;
  clearVoltages();
  endDataWrite();
  updateHost();
}



void
MeshEngine::deleteSimulationData()
{
  startDataWrite();
  currentIteration_ = 0;
  releaseResources();
  meshSize_ = MeshSize(0,0,0);
  foreach(ProbeMeshItem* probe, meshProbes_){
    probe->clearData();
  }
  endDataWrite();
  emit voltageDataChanged();
}




void
MeshEngine::clearVoltages()
{
  blockSignals(true);
  MeshPoint meshPoint;
  for(meshPoint.rz() = 0; meshPoint.rz() < meshSize().depth(); ++meshPoint.rz()){
    for(meshPoint.ry() = 0; meshPoint.ry() < meshSize().height(); ++meshPoint.ry()){
      for(meshPoint.rx() = 0; meshPoint.rx() < meshSize().width(); ++meshPoint.rx()){
        for(int i = 1; i <= nodeVoltageCount_; ++i){
          setNodeVoltage(meshPoint,i,0);
        }
      }
    }
  }
  blockSignals(false);
}


//-----------------------------------------------------------------------------
// PROTECTED METHODS
//-----------------------------------------------------------------------------



void
MeshEngine::startDataWrite()
{
  meshEngineMutex_.lockForWrite();
}



void
MeshEngine::endDataWrite()
{
  meshEngineMutex_.unlock();
}



void
MeshEngine::startDataRead()
{
  meshEngineMutex_.lockForRead();
}



void
MeshEngine::endDataRead()
{
  meshEngineMutex_.unlock();
}



//-----------------------------------------------------------------------------
// PRIVATE METHODS
//-----------------------------------------------------------------------------



void MeshEngine::analyzeMeshItems(QList<MeshItem*>& meshItems)
{
//  Point3D<float> bottomBackLeft = meshItems.first()->pos();
//  Point3D<float> topFrontRight = bottomBackLeft;
//  Point3D<float> itemPos, itemFarPos;
//  foreach(MeshItem* item, meshItems){
//    itemPos = item->pos();
//    itemFarPos = item->pos()+item->size().toPoint3D();
//    bottomBackLeft = min(bottomBackLeft,min(itemPos,itemFarPos));
//    topFrontRight = max(topFrontRight,max(itemPos,itemFarPos));
//  }
  meshSize_ = MeshSize();
  meshOrigin_ = Point3D<float>();

  Point3D<float> bottomBackLeft = getBottomBackLeft(meshItems);
  Point3D<float> topFrontRight = getTopFrontRight(meshItems);
  meshOrigin_ = bottomBackLeft;
  MeshPoint meshTopFrontRight = discretizePoint(topFrontRight);
  meshSize_ = MeshSize(meshTopFrontRight);
  meshSize_ = Max(meshSize_,MeshSize(2,2,2));
}



void MeshEngine::setOuterMeshBoundary()
{
  MeshPoint point;
  MeshDirection boundaryDirection;

  // SET BOUNDARYS FOR TOP AND BOTTOM OF MESH
  if(meshSize_.height() > 0){
    for(point.ry() = 0; point.ry() < meshSize_.height(); point.ry() += meshSize_.height()-1){
      if(point.ry() == 0){
        boundaryDirection = PY;
      }
      if(point.ry() == meshSize_.height()-1){
        boundaryDirection = NY;
      }
      for(point.rz() = 0; point.rz() < meshSize_.depth(); point.rz() ++){
        for(point.rx() = 0; point.rx() < meshSize_.width(); point.rx() ++){
          setNodeBoundary(point,boundaryDirection,0.0);
        }
      }
    }
  }

  // SET BOUNDARIES FOR SIDES OF MESH
  if(meshSize_.width() > 0){
    for(point.rx() = 0; point.rx() < meshSize_.width(); point.rx() += meshSize_.width()-1){
      if(point.rx() == 0){
        boundaryDirection = NX;
      }
      if(point.rx() == meshSize_.width()-1){
        boundaryDirection = PX;
      }
      for(point.rz() = 0; point.rz() < meshSize_.depth(); point.rz() ++){
        for(point.ry() = 0; point.ry() < meshSize_.height(); point.ry() ++){
          setNodeBoundary(point,boundaryDirection,0.0);
        }
      }
    }
  }

  // SET BOUNDARIES FOR FRONT AND BACK OF MESH
//  if(meshSize_.depth() > 1){
//    for(point.rz() = 0; point.rz() < meshSize_.depth(); point.rz() += meshSize_.depth()-1){
//      if(point.rz() == 0){
//        boundaryDirection = PZ;
//      }
//      if(point.rz() == meshSize_.depth()-1){
//        boundaryDirection = NZ;
//      }
//      for(point.ry() = 0; point.ry() < meshSize_.height(); point.ry() ++){
//        for(point.rx() = 0; point.rx() < meshSize_.width(); point.rx() ++){
//          setNodeBoundary(point,boundaryDirection,0);
//        }
//      }
//    }
//  }
}



//-----------------------------------------------------------------------------
// VIRTUAL PUBLIC METHODS
//-----------------------------------------------------------------------------



void
MeshEngine::setNodeConductance(const MeshPoint &point,
                               const float conductance)
{
  QVector<float> tmpVec;
  tmpVec.append(conductance);
  setNodeConductance(point,tmpVec);
}



void
MeshEngine::setNodePermeability(const MeshPoint &point,
                                const float permeability)
{
  QVector<float> tmpVec;
  tmpVec.append(permeability);
  setNodeConductance(point,tmpVec);
}



void
MeshEngine::setNodePermittity(const MeshPoint &point,
                              const float permittivity)
{
  QVector<float> tmpVec;
  tmpVec.append(permittivity);
  setNodeConductance(point,tmpVec);
}



void
MeshEngine::createSimulation()
{
  if(isCompatible()){
    meshMutex_.lockForRead();
    QList<MeshItem*> meshItems = mesh_->items();
    meshMutex_.unlock();
    if( meshItems.count()>0 ){
      meshProbes_.clear();
      meshSources_.clear();
      blockSignals(true);
      startDataWrite();
      analyzeMeshItems(meshItems);
      releaseResources();
      allocateResources( meshSize_ );
      initializeResources( meshSize_ );
      setOuterMeshBoundary();
      foreach( MeshItem* meshItem, meshItems){
        addMeshItem( meshItem );
      }

      endDataWrite();
      blockSignals(false);
      emit voltageDataChanged();
    }
  }
  else{
    QMessageBox(QMessageBox::Critical,
                QString(tr("Incompatible Computational Engine")),
                QString(tr("Your computer is incompatible with the requested computational engine.")),
                QMessageBox::Ok).exec();
  }
}



void
MeshEngine::simulateIteration(int iteration)
{
  currentIteration_ = iteration;
}




void
MeshEngine::updateHost()
{
  emit voltageDataChanged();
}



//-----------------------------------------------------------------------------
// VIRTUAL PROTECTED METHODS
//-----------------------------------------------------------------------------






