/*******************************************************************************
**  FILE:
**  DESCRIPTION:
**
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include <QObject>
#include <QtGui>
#include "../inc/oclmanager.h"



//-----------------------------------------------------------------------------
// CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------


OCLManager::OCLManager(QString initKernelFilePath, QObject *parent)
  : QObject(parent)
  , kernelFilePath(initKernelFilePath)
  , contextGL(NULL)
  , context(NULL)
  , profilingEnabled (false)
  , scatterKernelRunTime (0)
  , interchangeKernelRunTime (0)
  , exciteKernelRunTime (0)
  , sampleKernelRunTime (0)
{
    hostCompatible = setupOpenCL();
}



OCLManager::OCLManager(const OCLManager &other)
{
    hostCompatible = other.hostCompatible;
    kernelFilePath = other.kernelFilePath;

    contextGL     = other.contextGL;
    context       = other.context;
    program       = other.program;
    commandQueue  = other.commandQueue;

    exciteKernel              = other.exciteKernel;
    sampleKernel              = other.sampleKernel;
    scatterKernel             = other.scatterKernel;
    impulseInterchangeKernel  = other.impulseInterchangeKernel;

    boundaryTypeTable   = other.boundaryTypeTable;
    voltages            = other.voltages;
    boundaries          = other.boundaries;
    properties          = other.properties;
    probeInformation_   = other.probeInformation_;
    probeData_          = other.probeData_;
    sourceInformation_  = other.sourceInformation_;
    sourceData_         = other.sourceData_;

    numSources_         = other.numSources_;
    numProbes_          = other.numProbes_;
}



OCLManager&
OCLManager::operator =(const OCLManager &other)
{
  if(&other!=this){
    hostCompatible = other.hostCompatible;
    kernelFilePath = other.kernelFilePath;

    contextGL = other.contextGL;
    context = other.context;
    program = other.program;
    commandQueue = other.commandQueue;

    exciteKernel = other.exciteKernel;
    sampleKernel = other.sampleKernel;
    scatterKernel = other.scatterKernel;
    impulseInterchangeKernel = other.impulseInterchangeKernel;

    boundaryTypeTable = other.boundaryTypeTable;
    voltages = other.voltages;
    boundaries = other.boundaries;
    properties = other.properties;
    probeInformation_ = other.probeInformation_;
    probeData_ = other.probeData_;
    sourceInformation_ = other.sourceInformation_;
    sourceData_ = other.sourceData_;

    numSources_ = other.numSources_;
    numProbes_ = other.numProbes_;
  }
  return *this;
}



OCLManager::~OCLManager()
{
    if(contextGL)
        delete contextGL;
    if(context)
        delete context;
    releaseBuffers();
}



//-----------------------------------------------------------------------------
// PUBLIC METHODS
//-----------------------------------------------------------------------------


const bool OCLManager::isHostCompatible() const
{
  return hostCompatible;
}

// MEMORY MANAGEMENT

void
OCLManager::copyBoundaryTypeTableToDevice(float *data, size_t size){
  if(!boundaryTypeTable.isNull()){
    boundaryTypeTable.release();
  }
  boundaryTypeTable = context->createVector<float>(size,QCLMemoryObject::ReadOnly);
  boundaryTypeTable.write(data,size);
}



void
OCLManager::copyBoundariesToDevice(unsigned short *data, size_t size)
{
  if(!boundaries.isNull()){
    boundaries.release();
  }
  boundaries = context->createVector<unsigned short>(size,QCLMemoryObject::ReadOnly);
  boundaries.write(data,size);
}



void
OCLManager::copyPropertiesToDevice(float *data, size_t size)
{
  if(!properties.isNull()){
    properties.release();
  }
  properties = context->createVector<float>(size,QCLMemoryObject::ReadOnly);
  properties.write(data,size);
}


void
OCLManager::copyProbesToDevice(int numProbes,
                                    QVector<unsigned int> &probeInfoVector,
                                    int probeBufferSize)
{
  numProbes_ = numProbes;
  if(!probeInformation_.isNull()){
    probeInformation_.release();
  }
  if(!probeData_.isNull()){
    probeData_.release();
  }
  if(numProbes>0){
    probeInformation_ = context->createVector<unsigned int>(probeInfoVector.size(),QCLMemoryObject::ReadOnly);
    probeInformation_.write(probeInfoVector);
    probeData_ = context->createVector<float>(numProbes_*probeBufferSize,QCLMemoryObject::WriteOnly);
  }
}


void
OCLManager::copySourcesToDevice(int numSources,
                                     QVector<unsigned int> &sourceInfoVector,
                                     QVector<float> &sourceDataVector)
{
  numSources_ = numSources;
  if(!sourceInformation_.isNull()){
    sourceInformation_.release();
  }
  if(!sourceData_.isNull()){
    sourceData_.release();
  }
  if(numSources>0){
    sourceInformation_ = context->createVector<unsigned int>(sourceInfoVector.size(),QCLMemoryObject::ReadOnly);
    sourceInformation_.write(sourceInfoVector);
    sourceData_ = context->createVector<float>(sourceDataVector.size(),QCLMemoryObject::ReadOnly);
    sourceData_.write(sourceDataVector);
  }
}


void
OCLManager::createVoltageBuffer(float* data,size_t size)
{
  if(!voltages.isNull()){
    voltages.release();
  }
  voltages = context->createVector<float>(size,QCLMemoryObject::ReadWrite);
  voltages.write(data,size);
}



void
OCLManager::copyVoltagesToHost(float *data, size_t size)
{
    voltages.read(data,size);
}



void
OCLManager::copyProbesToHost(float *data, size_t size)
{
    probeData_.read(data,size);
}



// SIMULATION MANAGEMENT

void
OCLManager::exciteMesh(int currentIteration)
{
  exciteKernel.setArg(5,currentIteration);
  QCLEvent exciteKernelEvent = exciteKernel.run();
  if(profilingEnabled){
    exciteKernelEvent.waitForFinished();
    exciteKernelRunTime += exciteKernelEvent.finishTime()-exciteKernelEvent.runTime();
  }
}


void
OCLManager::sampleMesh(int currentIteration)
{
  sampleKernel.setArg(6,currentIteration);
  QCLEvent sampleKernelEvent = sampleKernel.run();
  if(profilingEnabled){
    sampleKernelEvent.waitForFinished();
    sampleKernelRunTime += sampleKernelEvent.finishTime()-sampleKernelEvent.runTime();
  }

}


void
OCLManager::scatterImpulses()
{

  QCLEvent scatterKernelEvent = scatterKernel.run();
  if(profilingEnabled){
    scatterKernelEvent.waitForFinished();
    scatterKernelRunTime += scatterKernelEvent.finishTime()-scatterKernelEvent.runTime();
  }
}

void
OCLManager::impulseInterchange()
{
  QCLEvent impulseInterchangeKernelEvent = impulseInterchangeKernel.run();
  if(profilingEnabled){
    impulseInterchangeKernelEvent.waitForFinished();
    interchangeKernelRunTime += impulseInterchangeKernelEvent.finishTime() -
        impulseInterchangeKernelEvent.runTime();
  }
}


void
OCLManager::releaseBuffers(){
  boundaryTypeTable.release();
  voltages.release();
  boundaries.release();
  properties.release();
  probeInformation_.release();
  probeData_.release();
  sourceInformation_.release();
  sourceData_.release();
}


void
OCLManager::setProfilingEnabled(const bool enabled)
{
  if(hostCompatible){
    profilingEnabled = enabled;
    createCommandQueue();
  }
}


const bool
OCLManager::isProfilingEnabled() const
{
  return profilingEnabled;
}


const quint64
OCLManager::totalSamplingKernelRunTime() const
{
  return sampleKernelRunTime;
}


const quint64
OCLManager::totalExcitationKernelRunTime() const
{
  return exciteKernelRunTime;
}


const quint64
OCLManager::totalScatteringKernelRunTime() const
{
  return scatterKernelRunTime;
}


const quint64
OCLManager::totalInterchangeKernelRunTime() const
{
  return interchangeKernelRunTime;
}


//-----------------------------------------------------------------------------
// PRIVATE METHODS
//-----------------------------------------------------------------------------


bool
OCLManager::setupOpenCL()
{
  bool status;
  status = setupOpenCLContext();
  if(!status){
    qDebug("Failed to create OpenCL context.");
    return false;
  }else{
    status = createProgramFromFile(kernelFilePath);
    if(!status){
	    qDebug("Failed to create OpenCL program.");
	    return false;
    }else{
	    status = buildProgram();
	    if(!status){
        qDebug("Failed to build program OpenCL program.");
        return false;
	    }else{
        status = createKernels();
        if(!status){
          qDebug("Failed to build create OpenCL Kernel objects.");
          return false;
        }
	    }
    }
  }
  return true;
}


bool
OCLManager::setupOpenCLContext()
{
  //    contextGL = new QCLContextGL();
  //    // Try to create OpenCLGL context first and if that fails
  //    // then create a standard OpenCL context.
  //    if(!contextGL.create()){
  context = new QCLContext();
  if(!context->create()){
    delete context;
    return false;
  }
  return createCommandQueue();

  //	delete contextGL;
  //    }else{
  //	qDebug()<<"Supports CLGL:"<<contextGL.supportsObjectSharing();
  //    }
}


bool OCLManager::createCommandQueue()
{
  if(profilingEnabled){
    commandQueue = context->createCommandQueue(CL_QUEUE_PROFILING_ENABLE);
  }else{
    commandQueue = context->commandQueue();
  }
  if(!commandQueue.isNull()){
    context->setCommandQueue(commandQueue);
    return true;
  }else{
    return false;
  }
}


bool
OCLManager::createProgramFromFile(const QString fileName)
{
  QFile file;
  qDebug()<<"File '"<<fileName<<"' exists:"<<file.exists(fileName);
  if(file.exists(fileName)){
    program = context->createProgramFromSourceFile(fileName);
    if(!program.isNull()){
      return true;
    }else{
      return false;
    }
  }else{
    qDebug()<<"Could not find the file "<<fileName;
    return false;
  }
}


bool
OCLManager::buildProgram(){
  bool buildSuccess = program.build();
  if(!buildSuccess){
    qDebug()<<"Errors occured while building OpenCL program.";
    qDebug()<<program.log();
    qDebug("Invalid OpenCL Program.");
    return false;
  }else{
    return true;
  }
}


bool
OCLManager::createKernels(){
  // Excitation Kernel
  exciteKernel = program.createKernel(QString("excite"));
  if(exciteKernel.isNull()){
    qDebug()<<"Failed to create excitation kernel";
    return false;
  }

  // Sampling Kernel
  sampleKernel = program.createKernel(QString("sample"));
  if(sampleKernel.isNull()){
    qDebug()<<"Failed to create sampling kernel";
    return false;
  }

  // Scattering Kernel
  scatterKernel = program.createKernel(QString("scatter"));
  if(scatterKernel.isNull()){
    qDebug()<<"Failed to create scattering kernel";
    return false;
  }

  // Impulse Interchange Kernel
  impulseInterchangeKernel = program.createKernel(QString("impulseInterchange"));
  if(impulseInterchangeKernel.isNull()){
    qDebug()<<"Failed to create impulse interchange kernel";
    return false;
  }

  return true;
}



QCLWorkSize
OCLManager::getGlobalWorkSize(const MeshSize &meshSize, QCLWorkSize localSize)
{
  // 3D case if the volume is larger than the area
  if(meshSize.volume()>meshSize.area()){
    size_t globalWidth = ceil((double)meshSize.width()/localSize.width())*localSize.width();
    size_t globalHeight = ceil((double)meshSize.height()/localSize.height())*localSize.height();
    size_t globalDepth = ceil((double)meshSize.depth()/localSize.depth())*localSize.depth();
    QCLWorkSize globalSize(globalWidth,globalHeight,globalDepth);
    return globalSize;
  }else{
    size_t globalWidth = ceil((double)meshSize.width()/localSize.width())*localSize.width();
    size_t globalHeight = ceil((double)meshSize.height()/localSize.height())*localSize.height();
    if(meshSize.width()<localSize.width()){
      globalWidth = localSize.width();
    }
    if(meshSize.height()<localSize.height()){
      globalHeight = localSize.height();
    }
    QCLWorkSize globalSize(globalWidth,globalHeight);
    return globalSize;
  }
}




//-----------------------------------------------------------------------------
// PUBLIC SLOTS
//-----------------------------------------------------------------------------



void OCLManager::oclFinish()
{
  context->finish();
}


void
OCLManager::wait(int mSecs)
{
  QTime waitTime = QTime::currentTime().addMSecs(mSecs);
  while(QTime::currentTime()<waitTime);
}



void
OCLManager::setupKernels(const MeshSize meshSize,
                         size_t workItemSharedMemSize,
                         const MeshSize localWorkSize)
{
    setupScatteringKernel(meshSize,workItemSharedMemSize,localWorkSize);
    setupImpulseInterchangeKernel(meshSize,workItemSharedMemSize,localWorkSize);
    setupExcitationKernel(meshSize,workItemSharedMemSize,localWorkSize);
    //setupSamplingKernel(meshSize,workItemSharedMemSize,localWorkSize);
    wait(50);
}



void
OCLManager::setupScatteringKernel(const MeshSize meshSize,
                                  size_t workItemSharedMemSize,
                                  const MeshSize localWorkSize)
{
//    cl_uint width = meshSize.width();
//    cl_uint height = meshSize.height();
//    cl_uint depth = meshSize.depth();

    QCLWorkSize localSize,globalSize;

    if(localWorkSize.area() == 0){
        if(localWorkSize.volume()>localWorkSize.area()){
            localSize = scatterKernel.bestLocalWorkSizeImage3D();
        }else{
            localSize = scatterKernel.bestLocalWorkSizeImage2D();
        }
    }else{
        localSize = QCLWorkSize(localWorkSize.width(),localWorkSize.height(),localWorkSize.depth());
    }
    globalSize = getGlobalWorkSize(meshSize,localSize);

    scatterKernel.setGlobalWorkSize(globalSize);
    scatterKernel.setLocalWorkSize(localSize);


    // For some odd reason if I try to use a variable to declare the size of the local submesh
    // there is an error
    size_t subMeshSize = workItemSharedMemSize*localSize.width()*localSize.height();



    scatterKernel.setArg(0,voltages);               // Voltage Buffer
    scatterKernel.setArg(1,properties);             // Property buffer
    scatterKernel.setArg(2,meshSize.width());                  // Mesh Width
    scatterKernel.setArg(3,meshSize.height());                 // Mesh Height
    scatterKernel.setArg(4,meshSize.depth());                  // Mesh Depth
    scatterKernel.setArg(5,NULL,subMeshSize);              // Shared Memory (5 voltages x 8 x 8 nodes 320 floats)
}



void
OCLManager::setupImpulseInterchangeKernel(const MeshSize meshSize,
                                          size_t workItemSharedMemSize,
                                          const MeshSize localWorkSize)
{

    QCLWorkSize localSize,globalSize;

    if(localWorkSize.area()==0){
        if(localWorkSize.volume()>localWorkSize.area()){
            localSize = impulseInterchangeKernel.bestLocalWorkSizeImage3D();
        }else{
            localSize = impulseInterchangeKernel.bestLocalWorkSizeImage2D();
        }
    }else{
        localSize = QCLWorkSize(localWorkSize.width(),localWorkSize.height(),localWorkSize.depth());
    }
    globalSize = getGlobalWorkSize(meshSize,localSize);

    impulseInterchangeKernel.setGlobalWorkSize(globalSize);
    impulseInterchangeKernel.setLocalWorkSize(localSize);

    // For some odd reason if I try to use a variable to declare the size of the local submesh
    // there is an error
    size_t subMeshSize = workItemSharedMemSize*localSize.width()*localSize.height();


    impulseInterchangeKernel.setArg(0,voltages);		// Voltage Buffer
    impulseInterchangeKernel.setArg(1,boundaries);		// Boundary Code buffer
    impulseInterchangeKernel.setArg(2,boundaryTypeTable);	// Bounary Type Table buffer
    impulseInterchangeKernel.setArg(3,meshSize.width());	// Mesh Width
    impulseInterchangeKernel.setArg(4,meshSize.height());	// Mesh Height
    impulseInterchangeKernel.setArg(5,meshSize.depth());			// Mesh Depth
    impulseInterchangeKernel.setArg(6,NULL,subMeshSize);	// Shared Memory (5 voltages x 8 x 8 nodes + 1 for buffer protection 321 floats)
}



void
OCLManager::setupExcitationKernel(const MeshSize meshSize,
                                  size_t workItemSharedMemSize,
                                  const MeshSize localWorkSize)
{
  if(numSources_>0){
    QCLWorkSize localSize,globalSize;

    //    if(localWorkSize.area()==0){
    //        if(localWorkSize.volume()>localWorkSize.area()){
    //	    localSize = exciteKernel.bestLocalWorkSizeImage3D();
    //        }else{
    //	    localSize = exciteKernel.bestLocalWorkSizeImage2D();
    //        }
    //    }else{
    //        localSize = QCLWorkSize(localWorkSize.width(),localWorkSize.height(),localWorkSize.depth());
    //    }
    //    globalSize = getGlobalWorkSize(meshSize,localSize);

    localSize = QCLWorkSize(numSources_,1,1);
    globalSize = localSize;

    exciteKernel.setGlobalWorkSize(globalSize);
    exciteKernel.setLocalWorkSize(localSize);

    exciteKernel.setArg(0,voltages);               // Voltage Buffer
    exciteKernel.setArg(1,sourceInformation_);      // source information
    exciteKernel.setArg(2,sourceData_);             // Source data
    exciteKernel.setArg(3,numSources_);             // Number of sources
    exciteKernel.setArg(4,sourceInformation_.size()/numSources_);
    exciteKernel.setArg(5,0);
    exciteKernel.setArg(6,meshSize.width());
    exciteKernel.setArg(7,meshSize.height());
    exciteKernel.setArg(8,meshSize.depth());
  }
}



void
OCLManager::setupSamplingKernel(const MeshSize meshSize,
                                size_t workItemSharedMemSize,
                                const MeshSize localWorkSize)
{
  if(numProbes_>0){
    QCLWorkSize localSize,globalSize;

    //    if(localWorkSize.area()==0){
    //        if(localWorkSize.volume()>localWorkSize.area()){
    //	    localSize = exciteKernel.bestLocalWorkSizeImage3D();
    //        }else{
    //	    localSize = exciteKernel.bestLocalWorkSizeImage2D();
    //        }
    //    }else{
    //        localSize = QCLWorkSize(localWorkSize.width(),localWorkSize.height(),localWorkSize.depth());
    //    }
    //    globalSize = getGlobalWorkSize(meshSize,localSize);

    localSize = QCLWorkSize(numProbes_,1,1);
    globalSize = localSize;

    sampleKernel.setGlobalWorkSize(globalSize);
    sampleKernel.setLocalWorkSize(localSize);

    sampleKernel.setArg(0,voltages);               // Voltage Buffer
    sampleKernel.setArg(1,probeInformation_);      // Probe information
    sampleKernel.setArg(2,probeData_);             // Sample Buffer
    sampleKernel.setArg(3,probeInformation_.size()/numProbes_); // Stride of probe info vector
    sampleKernel.setArg(4,probeData_.size()/numProbes_);  // Buffer Size
    sampleKernel.setArg(5,properties);                    // Node properties
    sampleKernel.setArg(6,0);                             // Iteration
    sampleKernel.setArg(7,meshSize.width());
    sampleKernel.setArg(8,meshSize.height());
    sampleKernel.setArg(9,meshSize.depth());
  }
}
