/*******************************************************************************
**  FILE:
**  DESCRIPTION:
**
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include "inc/tlmengine2d.h"
#include "inc/meshitems.h"


//-----------------------------------------------------------------------------
// CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------


TLMEngine2D::TLMEngine2D(Mesh* initMesh)
  : MeshEngine      (initMesh)
  , nodeVoltages_   (0)
  , nodeBoundaries_ (0)
  , nodeProperties_ (0)
  , numBoundaryTypes_(0)
{
  nodeVoltageCount_   = 5;
  nodeBoundaryCount_  = 4;
  nodePropertyCount_  = 2;
  dimensions_         = 2;
  engineName_         = tr("2D TLM Shunt Node");
  setCompatibility(true);
}



TLMEngine2D::TLMEngine2D(const TLMEngine2D &other)
  : MeshEngine      (other)

{
  nodeVoltageCount_   = other.nodeVoltageCount_;
  nodeBoundaryCount_  = other.nodeBoundaryCount_;
  nodePropertyCount_  = other.nodePropertyCount_;
  dimensions_         = other.dimensions_;
  engineName_         = other.engineName_;
  setCompatibility(true);
}



TLMEngine2D::~TLMEngine2D()
{

}



//-----------------------------------------------------------------------------
// PUBLIC METHODS
//-----------------------------------------------------------------------------


// Operator overloads

TLMEngine2D&
TLMEngine2D::operator =(const TLMEngine2D& other)
{
  if(this!=&other){
    MeshEngine::operator =(other);
  }
  return *this;
}



MeshEngine*
TLMEngine2D::clone() const
{
    return new TLMEngine2D(*this);
}



void
TLMEngine2D::setNodeBoundary(const MeshPoint &point,
                            const MeshEngine::MeshDirection direction,
                            const float value)
{
  int boundary = -1;
  int otherBoundary;
  MeshPoint otherPoint;
  USHORT newBoundID = addBoundaryValue(value);
  switch(direction){
  case MeshEngine::NY:
    boundary = 1;
    otherBoundary = 3;
    otherPoint = point + MeshPoint(0,1,0);
    break;
  case MeshEngine::PY:
    boundary = 3;
    otherBoundary = 1;
    otherPoint = point - MeshPoint(0,1,0);
    break;
  case MeshEngine::NX:
    boundary = 2;
    otherBoundary = 4;
    otherPoint = point - MeshPoint(1,0,0);
    break;
  case MeshEngine::PX:
    boundary = 4;
    otherBoundary = 2;
    otherPoint = point + MeshPoint(1,0,0);
    break;
  }
  if(boundary != -1){
    USHORT newBoundID = addBoundaryValue(value);
    MeshPoint meshSizePoint = meshSize_.toPoint3D()+MeshPoint(0,0,1);
    MeshPoint zeroPoint = MeshPoint(0,0,0);
//    if(boundary == 2 || otherBoundary == 2){
//      qDebug()<<point<<" | "<<boundary<<" | "<<otherPoint<<" | "<<otherBoundary<<" | "<<newBoundID<<"| Value:"<<value;
//    }
    if(point < meshSizePoint && point >= zeroPoint){
      nodeBoundaries_[getMemoryIndex(point,boundary)] = newBoundID;
    }
    if(otherPoint < meshSizePoint && otherPoint >= zeroPoint){
      nodeBoundaries_[getMemoryIndex(otherPoint,otherBoundary)] = newBoundID;
    }
  }
}



void
TLMEngine2D::setNodeConductance(const MeshPoint &point,
                                     const QVector<float> conductance)
{
  nodeProperties_[getMemoryIndex(point,2)] = conductance.first();
}



void
TLMEngine2D::setNodePermeability(const MeshPoint &point,
                                      const QVector<float> permeability)
{
 // This mesh doesn't support permeability yet
}



void
TLMEngine2D::setNodePermittity(const MeshPoint &point,
                               const QVector<float> permittivity)
{
  nodeProperties_[getMemoryIndex(point,1)] = permittivity.first();
}



void
TLMEngine2D::setNodeVoltage(const MeshPoint &point,
                                const int voltage,
                                const float value)
{
  nodeVoltages_[getMemoryIndex(point,voltage)]=value;
  emit voltageDataChanged();
}



float&
TLMEngine2D::getNodeVoltage(const MeshPoint &point, const int voltage)
{
  return nodeVoltages_[getMemoryIndex(point,voltage)];
}



void
TLMEngine2D::simulateIteration(int iteration)
{
  if(isCompatible()){
    MeshEngine::simulateIteration(iteration);
    startDataWrite();
      exciteMesh(iteration);
      scatterImpulses();
      impulseInterchange();
    endDataWrite();
    startDataRead();
      sampleMesh(iteration);
    endDataRead();
  }
}



void
TLMEngine2D::updateHost()
{
  MeshEngine::updateHost();
}


void
TLMEngine2D::createSimulation()
{
  MeshEngine::createSimulation();
  meshSize_.setDepth(1);
}


//-----------------------------------------------------------------------------
// PROTECTED METHODS
//-----------------------------------------------------------------------------


void
TLMEngine2D::allocateResources(const MeshSize &meshSize)
{
//  qDebug()<<"Allocating Resources...";
  try{
    nodeVoltages_ = new float[meshSize.area()*nodeVoltageCount_];
    nodeBoundaries_ = new USHORT[meshSize.area()*nodeBoundaryCount_];
    nodeProperties_ = new float[meshSize.area()*nodePropertyCount_];
  }
  catch(const std::exception &e){
    qDebug()<<"Error:"<<e.what();
    setCompatibility(false);
  }
//  qDebug()<<"Resources Allocated.";
}



void
TLMEngine2D::initializeResources(const MeshSize &meshSize)
{
//  qDebug()<<"Initializing Resources...";
  for(int row = 0; row < meshSize.height(); ++row){
    for(int col = 0; col < meshSize.width(); ++col){
      int nodePosition = row*meshSize.width()+col;
      for(int v = 0; v < nodeVoltageCount_; ++v){
        nodeVoltages_[v*meshSize.area()+nodePosition] = 0;
      }

      float defaultProperties[2] = {0.0, 0.0}; // Admittance, Conductance
      for(int p = 0; p < nodePropertyCount_; ++p){
        nodeProperties_[p*meshSize.area()+nodePosition] = defaultProperties[p];
      }

      for(int b = 0; b < nodeBoundaryCount_; ++b){
        nodeBoundaries_[b*meshSize.area()+nodePosition] = 0;
      }
      numBoundaryTypes_ = 0;
    }
  }
  emit voltageDataChanged();
//  qDebug()<<"Resources Initialized.";
}



void
TLMEngine2D::releaseResources()
{
//  qDebug()<<"Releasing Resources...";
  if(nodeVoltages_ != 0)
    delete[] nodeVoltages_;
  if(nodeBoundaries_ != 0)
    delete[] nodeBoundaries_;
  if(nodeProperties_ != 0)
    delete[] nodeProperties_;

  nodeVoltages_   = 0;
  nodeProperties_ = 0;
  nodeBoundaries_ = 0;
//  qDebug()<<"Resources Released.";
}



void
TLMEngine2D::scatterImpulses()
{
  // INHOMOGENEOUS SHUNT NODE SCATTERING PROCEDURE
  float a, Ys;
  int v1,v2,v3,v4,v5;
  MeshPoint point = MeshPoint(0,0,0);
  for (point.ry() = 0; point.ry() < meshSize_.height(); ++point.ry())
  {
    for (point.rx() = 0; point.rx() < meshSize_.width(); ++point.rx())
    {
      v1 = getMemoryIndex(point,1);
      v2 = getMemoryIndex(point,2);
      v3 = getMemoryIndex(point,3);
      v4 = getMemoryIndex(point,4);
      v5 = getMemoryIndex(point,5);
      Ys = nodeProperties_[getMemoryIndex(point,1)];

      // Calculate the common portion of the voltage scattering equation
      a=(nodeVoltages_[v1]+nodeVoltages_[v2]+nodeVoltages_[v3]
         +nodeVoltages_[v4]+nodeVoltages_[v5]*Ys)*2.0/
          (Ys + nodeProperties_[getMemoryIndex(point,2)] + 4);

      nodeVoltages_[v1]=a-nodeVoltages_[v1];
      nodeVoltages_[v2]=a-nodeVoltages_[v2];
      nodeVoltages_[v3]=a-nodeVoltages_[v3];
      nodeVoltages_[v4]=a-nodeVoltages_[v4];
      nodeVoltages_[v5]=a-nodeVoltages_[v5];
    }
  }
}



void
TLMEngine2D::impulseInterchange()
{
  // IMPULSE INTERCHANGE
  float reflection;
  USHORT boundaryCode;
  float tmpVoltage;
  int otherVoltageNumber;
  int voltageIndex, otherVoltageIndex;
  MeshPoint point = MeshPoint(0,0,0);
  MeshPoint nextPoint;
  for (point.ry() = 0; point.ry() < meshSize_.height(); ++point.ry())
  {
    for (point.rx() = 0; point.rx() < meshSize_.width(); ++point.rx())
    {
      nextPoint = point;
      for( int voltageNum = 1; voltageNum <= nodeBoundaryCount_; ++ voltageNum){
        voltageIndex = getMemoryIndex(point,voltageNum);
        boundaryCode = nodeBoundaries_[getMemoryIndex(point,voltageNum)];

        //  If there is a boundary then reflect the impulse back toward the node
        if(boundaryCode != 0){
          nodeVoltages_[voltageIndex] *= boundaryTypeTable_[boundaryCode];
        }
        //  If there is no boundary then interchange the impulse with
        //  the neighbouring node
        else{
          // Determine the node to interchange impulses with
          if(voltageNum == 1){
            if((point.cy() < meshSize_.height()-1)){
              nextPoint += MeshPoint(0,1,0);
              otherVoltageNumber = 3;
              flipVoltages(point,voltageNum,nextPoint,otherVoltageNumber);
            }else{
              nodeVoltages_[voltageIndex] = 0;
            }
          }
          else if(voltageNum == 2){
            if(point.cx() == 0){
              nodeVoltages_[voltageIndex] = 0;
            }
          }
          else if(voltageNum == 3){
            if(point.cy() == 0){
              nodeVoltages_[voltageIndex] = 0;
            }
          }
          else if(voltageNum == 4){
            if(point.cx() < meshSize_.width()-1){
              nextPoint = point + MeshPoint(1,0,0);
              otherVoltageNumber = 2;
              flipVoltages(point,voltageNum,nextPoint,otherVoltageNumber);
            }else{
              nodeVoltages_[voltageIndex] = 0;
            }
          }
        }
      } // end of voltage loop
    }   // end of column loop
  }     // end of row loop
}



void
TLMEngine2D::exciteMesh(int iteration)
{
  Point3D<float> position;
  MeshPoint meshPos;
  Size3D<float> size;

  float voltage;

  foreach(SourceMeshItem* source, meshSources_){
    if(source->getData().count() > iteration){
      meshPos = discretizePoint(source->pos());
      voltage = source->getData().at(iteration);
      for(int i = 1; i <= 4; ++i){
        nodeVoltages_[getMemoryIndex(meshPos,i)] += voltage;
      }
    }
  }
}



void
TLMEngine2D::sampleMesh(int iteration)
{
  if(meshProbes_.count()>0){
    MeshPoint probePos;
    float avgVoltage;
    float y0;
    foreach(ProbeMeshItem* probe, meshProbes_){
      avgVoltage = 0;
      probePos = discretizePoint(probe->pos());
      y0 = nodeProperties_[getMemoryIndex(probePos,1)];
      for(int voltage = 1; voltage <= 5; ++voltage){
        if(voltage < 5){
          avgVoltage += getNodeVoltage(probePos,voltage);
        }else{
          avgVoltage += y0*getNodeVoltage(probePos,voltage);
        }
      }
      avgVoltage *= 2.0/(y0 + 4.0);
      probe->appendData(avgVoltage);
    }
  }
}



const USHORT
TLMEngine2D::addBoundaryValue( float value )
{
  if(numBoundaryTypes_==256){
    qDebug()<<"Boundary Table full, cannot add boundary value";
    return 0;
  }
  int ID = getBoundaryID(value);
  if(ID==0){
    ID = numBoundaryTypes_;
    boundaryTypeTable_[ID] = value;
    numBoundaryTypes_++;
  }
  return ID;
}



const USHORT
TLMEngine2D::getBoundaryID(float value)
{
    USHORT ID = 0;
    for(int i=1;i<numBoundaryTypes_;++i){
        if(boundaryTypeTable_[i]==value){
            ID = i;
            break;
        }
    }
    return ID;
}



const int
TLMEngine2D::getMemoryIndex(const MeshPoint& point, int memberNumber) const
{
    int index = 0;
    index += (memberNumber-1)*meshSize_.area();
    index += point.cy()*meshSize_.width();
    index += point.cx();
    return index;
}




void
TLMEngine2D::flipVoltages(MeshPoint point1, int voltage1,
                          MeshPoint point2, int voltage2)
{
  int index1 = getMemoryIndex(point1,voltage1);
  int index2 = getMemoryIndex(point2,voltage2);
  float tmpVoltage = nodeVoltages_[index1];
  nodeVoltages_[index1] = nodeVoltages_[index2];
  nodeVoltages_[index2] = tmpVoltage;
}
