/*******************************************************************************
**  FILE:           
**  DESCRIPTION:    
**
**
**
**  AUTHOR:         Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:     Dr. Poman So
**
*******************************************************************************/

#include <QtGui>
#include "animatorglwidget.h"
#include "UI/Graphics/trackball.h"
#include "meshengine.h"
#include "UI/MeshSimulator/meshsimulator.h"

Size3D<int> s3di;
//-----------------------------------------------------------------------------
//  CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------

AnimatorGLWidget::AnimatorGLWidget(QWidget *parent)
  : GLDisplayWidget(parent)
  , meshEngine_ (0)
  , distanceExponent_ (0)
  , defaultDistance_ (0)
  , maxVoltage_(0)
{
}



AnimatorGLWidget::~AnimatorGLWidget()
{
}



//-----------------------------------------------------------------------------
//  PUBLIC METHODS
//-----------------------------------------------------------------------------


void
AnimatorGLWidget::setupViewingVolume(QSize viewSize) const
{
  gluPerspective(60,(float)(viewSize.width())/viewSize.height(),0.001,10000);
}



void
AnimatorGLWidget::initGL()
{
  GLDisplayWidget::initGL();
  glClearColor(0.0, 0.0, 0.0, 0.0);
}



void
AnimatorGLWidget::draw() const
{
  QMatrix4x4 view;
  view.rotate(trackBall_.rotation());
  glLoadIdentity();
  view(2, 3) -= defaultDistance_ * exp(distanceExponent_ / 1200.0f);
  multMatrix(view);
  if(surfacePoints_.size() != 0 && surfaceIndices_.size() != 0){
      glEnableClientState(GL_VERTEX_ARRAY);
      glEnableClientState(GL_COLOR_ARRAY);
      glColorPointer(3, GL_FLOAT, 0, surfaceColors_.data());
      glVertexPointer(3, GL_FLOAT, 0, surfacePoints_.data());
      glDrawElements(GL_LINE_STRIP, surfaceIndices_.size(),
                     GL_UNSIGNED_INT, surfaceIndices_.data());
      glDisableClientState(GL_COLOR_ARRAY);
      glDisableClientState(GL_VERTEX_ARRAY);
  }
}



//-----------------------------------------------------------------------------
//  PROTECTED METHODS
//-----------------------------------------------------------------------------



void
AnimatorGLWidget::mousePressEvent(QMouseEvent *event)
{
  QPointF mousePos = pixelPosToViewPos(event->posF());
  trackBall_.push(mousePos,trackBall_.rotation().conjugate());
  event->accept();
  updateGL();
}




void
AnimatorGLWidget::mouseMoveEvent(QMouseEvent *event)
{
  QPointF mousePos = pixelPosToViewPos(event->posF());
  trackBall_.move(mousePos,trackBall_.rotation().conjugate());
  event->accept();
  updateGL();
}



void
AnimatorGLWidget::mouseReleaseEvent(QMouseEvent *event)
{
  QPointF mousePos = pixelPosToViewPos(event->posF());
  trackBall_.release(mousePos,trackBall_.rotation().conjugate());
  event->accept();
  updateGL();
}



void
AnimatorGLWidget::wheelEvent(QWheelEvent *event)
{

  distanceExponent_ -= event->delta();
//  if (distanceExponent_ < -8 * 120)
//    distanceExponent_ = -8 * 120;
//  if (distanceExponent_ > 10 * 120)
//    distanceExponent_ = 10 * 120;
  event->accept();

  updateGL();
}



//-----------------------------------------------------------------------------
//  PRIVATE METHODS
//-----------------------------------------------------------------------------


QPointF AnimatorGLWidget::pixelPosToViewPos(const QPointF& p)
{
    return QPointF(2.0 * float(p.x()) / width() - 1.0,
                   1.0 - 2.0 * float(p.y()) / height());
}




void
AnimatorGLWidget::generateSurface()
{
  //meshEngineMutex_->lockForRead();
  if(meshSize_ != meshEngine_->meshSize()){
    meshSize_         = meshEngine_->meshSize();
    nodeSize_         = meshEngine_->meshResolution();
    distanceExponent_ = 0;
    defaultDistance_  = qMax(meshSize_.width()*nodeSize_.width(),
                             meshSize_.height()*nodeSize_.height());
    surfacePoints_.resize(meshSize_.area()*3);
    surfaceIndices_.resize(meshSize_.area()*2);
    surfaceColors_.resize(meshSize_.area()*3);
  }
  //meshEngineMutex_->unlock();

  float nodeWidth = nodeSize_.width();
  float nodeHeight = nodeSize_.height();
  // Generate points
  float xOffset,yOffset,zOffset;

  xOffset = (meshSize_.width()-1)*nodeWidth/2.0f;
  yOffset = (meshSize_.height()-1)*nodeHeight/2.0f;

  float curVoltage;
  float currentMaxVoltage = 0;
  //meshEngineMutex_->lockForRead();
  int nodeVoltageCount = meshEngine_->getNodeVoltageCount();
  QVector<GLfloat> averageVoltages;
  MeshPoint meshPoint;
  for(meshPoint.ry() = 0; meshPoint.ry() < meshSize_.height();++meshPoint.ry()){
    for(meshPoint.rx() = 0; meshPoint.rx() < meshSize_.width(); ++meshPoint.rx()){
      curVoltage = 0;
      for(int voltage = 1; voltage < nodeVoltageCount; ++voltage){
        curVoltage += meshEngine_->getNodeVoltage(meshPoint,voltage);
      }
      curVoltage /= nodeVoltageCount;
      averageVoltages<<curVoltage;
      currentMaxVoltage = qMax(qAbs(curVoltage),currentMaxVoltage);
    }
  }
  //meshEngineMutex_->unlock();
  maxVoltage_ = qMax(currentMaxVoltage,maxVoltage_);
  Point3D<GLfloat> glPoint;
  int pointIndex = 0;
  float maxHeight = 0.010;
  float heightNormalizer = (maxVoltage_!=0) ? maxHeight/maxVoltage_ : 1.0;

  float averageVoltage;
  for(meshPoint.ry() = 0; meshPoint.ry() < meshSize_.height();++meshPoint.ry()){
    for(meshPoint.rx() = 0; meshPoint.rx() < meshSize_.width(); ++meshPoint.rx()){
      averageVoltage = averageVoltages.at(meshPoint.ry()*meshSize_.width()+meshPoint.rx());
      glPoint.rx() = meshPoint.cx()*nodeWidth - xOffset;
      glPoint.ry() = yOffset - meshPoint.cy()*nodeHeight;
      glPoint.rz() = heightNormalizer*averageVoltage;

      surfaceColors_[pointIndex] = 1.0*qAbs(averageVoltage)/currentMaxVoltage;  // R
      surfacePoints_[pointIndex++] = glPoint.cx();  // X
      surfaceColors_[pointIndex] = 0.5;             // G
      surfacePoints_[pointIndex++] = glPoint.cy();  // Y
      surfaceColors_[pointIndex] = 0.5;             // B
      surfacePoints_[pointIndex++] = glPoint.cz();  // Z
      // Todo: Calculate normals and place in vector
    }
  }
  generateIndices();
}


void AnimatorGLWidget::generateIndices()
{
  GLuint index;
  int indexNum = 0;
  // First index loop for stitching the rows
  for(size_t y=0;y<meshSize_.height();++y){
    for(size_t x=0;x<meshSize_.width();++x){
      // if the row is even go from left to right
      // else go from right to left
      if(y%2==0){
        index = y*meshSize_.width()+x;
      }else{
        index = y*meshSize_.width() +(meshSize_.width()-(x+1));
      }
      surfaceIndices_[indexNum++] = index;
    }   // end of for x loop
  }   // end of for y loop

  // This check is determine which side of the mesh we end up on
  // in the first pass for stitching rows.
  if((meshSize_.height()-1)%2==0){
    // Second Index loop for stitching the columns
    for(int x=meshSize_.width()-1;x>=0;--x){
      for(int y=meshSize_.height()-1;y>=0;--y){
        if((x-(meshSize_.width()-1))%2==0){
          index = y*meshSize_.width()+ x;
        }else{
          index = (meshSize_.height()-(y+1))*meshSize_.width() + x;
        }
        surfaceIndices_[indexNum++] = index;
      }   // end of for x loop
    }   // end of for y loop
  }else{
    // Second Index loop for stitching the columns
    for(int x=0;x<meshSize_.width();++x){
      for(int y=meshSize_.height()-1;y>=0;--y){
        if(x%2==0){
          index = y*meshSize_.width()+ x;
        }else{
          index = (meshSize_.height()-(y+1))*meshSize_.width() + x;
        }
        surfaceIndices_[indexNum++] = index;
      }   // end of for x loop
    }   // end of for y loop
  }

}


//-----------------------------------------------------------------------------
//  PUBLIC SLOTS
//-----------------------------------------------------------------------------


void
AnimatorGLWidget::updateAnimation(MeshEngine *meshEngine)
{
  meshEngine_ = meshEngine;
  if(meshEngine_ != 0){
    generateSurface();
    updateGL();
  }
}



//void
//AnimatorGLWidget::meshEngineChanged(MeshEngine* meshEngine)
//{
//  meshEngine_ = meshEngine;
//  if(meshEngine_ != 0){
//    meshEngineMutex_ = &(meshEngine->getEngineMutex());
//  }else{
//    meshEngineMutex_ = 0;
//  }
//}
