/*******************************************************************************
**  FILE:           
**  DESCRIPTION:    
**
**
**
**  AUTHOR:         Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:     Dr. Poman So
**
*******************************************************************************/

#include <QtCore>
#include "dataset.h"
#include "datatransform.h"
#include "dataviewer.h"
#include "mesh.h"




/******************************
**
**       DATA TRANSFORM
**
******************************/


//-----------------------------------------------------------------------------
// CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------


DataTransform::DataTransform(DataViewer *dataViewer)
  : QObject(0)
  , dataViewer_ (dataViewer)
{
  #ifndef QT_NO_CONCURRENT
  connect(&transformWatcher,SIGNAL(finished()),this,SLOT(transformationFinished()));
  #endif
  displayText_ = tr("f(t)");
  toolTipText_ = tr("Time Domain");
  whatsThisText_ = tr("Time domain representation of signal.");
  meshResolution_ = dataViewer->mesh()->getMeshResolution();
}


DataTransform::~DataTransform()
{

}


//-----------------------------------------------------------------------------
// PUBLIC METHODS
//-----------------------------------------------------------------------------


void DataTransform::transformData(const DataSet &srcData, DataSet &destData)
{
  destData.setData();
#ifndef QT_NO_CONCURRENT
  QFuture<void> future = QtConcurrent::run(this,&DataTransform::computeNewData,
                                           &srcData,&destData);
  transformWatcher.setFuture(future);
#else
  computeNewData(&srcData,&destData);
  transformationFinished();
#endif
}



// GETTERS


QVariant
DataTransform::getOptionValue(const QString &optionName) const
{
  if( options_.contains( optionName ) ){
    return options_[ optionName ];
  }
  else{
    return QVariant(0);
  }
}



void
DataTransform::setOptionValue(const QString &optionName, const QVariant &value)
{
  if(options_.contains(optionName)){
    options_[optionName] = value;
  }
}



const QString
DataTransform::getOptionText(const QString &optionName) const
{
  if( options_.contains( optionName ) ){
    return optionText_[ optionName ];
  }
  else{
    return QString();
  }
}



const QMap<QString,QVariant>&
DataTransform::getOptionMap() const
{
  return options_;
}



const QMap<QString,QString>&
DataTransform::getOptionTextMap() const
{
  return optionText_;
}



const QString&
DataTransform::getDisplayText() const
{
  return displayText_;
}



const QString&
DataTransform::getWhatsThis() const
{
  return whatsThisText_;
}



const QString&
DataTransform::getToolTip() const
{
  return toolTipText_;
}


bool
DataTransform::hasOptions() const
{
  return (options_.count() > 0);
}



//-----------------------------------------------------------------------------
// PROTECTED METHODS
//-----------------------------------------------------------------------------


void
DataTransform::addOption(const QString &optionName,
                         const QString &optionTitle,
                         const QVariant optionValue)
{
  if( options_.contains( optionName ) ){
    options_[optionName] = optionValue;
  }
  else{
    options_.insert(optionName,optionValue);

  }
  if( optionText_.contains( optionName ) ){
    optionText_[optionName] = optionTitle;
  }
  else{
    optionText_.insert(optionName,optionTitle);
  }
}



void
DataTransform::removeOption(const QString &optionName)
{
  if( options_.contains( optionName ) ){
    options_.remove(optionName);
    optionText_.remove(optionName);
  }
}



void
DataTransform::computeNewData(const DataSet* srcData, DataSet* destData)
{
  *destData = *srcData;
  destData->analyzeData();
  destData->setTitle(tr("Time Domain"));
  destData->setXAxisTitle(tr("Sample (n)"));
  destData->setYAxisTitle(tr("Amplitude"));
}



//-----------------------------------------------------------------------------
// PRIVATE SLOTS
//-----------------------------------------------------------------------------


void
DataTransform::transformationFinished()
{
  emit dataTransformed();
}



/******************************
**
**    DATA TRANSFORM DIALOG
**
******************************/


#define QVARIANT_FLOAT   135

DataTransformDialog::DataTransformDialog(DataTransform* parent)
  : QDialog(qobject_cast<QWidget*>(parent),Qt::Dialog)
  , parent_(parent)
  , mainLayout_(new QFormLayout(this))
  , options_ ( parent_->getOptionMap())
  , optionText_( parent_->getOptionTextMap())
{
  setModal(true);
  createDialog();
  setLayout(mainLayout_);
}



void DataTransformDialog::createDialog()
{
  QMapIterator<QString, QString> mapIterator(optionText_);
  QWidget * label;
  QWidget* widget;
  QVariant optionData;
  QString optionKey;

  while(mapIterator.hasNext()){
    mapIterator.next();
    optionKey = mapIterator.key();
    label = new QLabel(mapIterator.value());
    optionData = options_.value(optionKey);
    QVariant::Type dataType = optionData.type();
    if(dataType == QVariant::Bool){
      widget = new QCheckBox;
      QCheckBox* checkBox = qobject_cast<QCheckBox*>(widget);
      checkBox->setChecked(optionData.value<bool>());
    }
    else{
      widget = new QLineEdit;
      QLineEdit* lineEdit = qobject_cast<QLineEdit*>(widget);
      if(dataType == QVARIANT_FLOAT){ // Equivilent to QVariant::float since it doesn't exist
        lineEdit->setText(QString("%1").arg(optionData.value<float>()));
        lineEdit->setValidator(new QDoubleValidator(this));
      }
      else if(dataType == QVariant::Int){
        lineEdit->setText(QString("%1").arg(optionData.value<int>()));
        lineEdit->setValidator(new QIntValidator(this));
      }
    }
    widget->setObjectName(optionKey);
    mainLayout_->addRow(label,widget);
  }
  buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
  connect(buttonBox,SIGNAL(accepted()),this,SLOT(accept()));
  connect(buttonBox,SIGNAL(rejected()),this,SLOT(reject()));
  mainLayout_->addWidget(buttonBox);
}


void DataTransformDialog::accept()
{
  QMapIterator<QString, QVariant> mapIterator(options_);
  QWidget* widget;
  QVariant::Type dataType;
  QVariant optionData;
  QVariant newData;
  QString optionKey;
  while(mapIterator.hasNext()){
    mapIterator.next();
    optionKey = mapIterator.key();
    optionData = mapIterator.value();

    dataType = optionData.type();
    widget = findChild<QWidget *>(optionKey);
    if(widget != 0){
      if(dataType == QVariant::Bool){
        QCheckBox* checkBox = qobject_cast<QCheckBox*>(widget);
        newData.setValue(checkBox->isChecked());
      }
      else{
        QLineEdit* lineEdit = qobject_cast<QLineEdit*>(widget);
        if(dataType == QVARIANT_FLOAT){
          newData.setValue(lineEdit->text().toFloat());
        }
        else if(dataType == QVariant::Int){
          newData.setValue(lineEdit->text().toInt());
        }
      }
      parent_->setOptionValue(optionKey,newData);
    }
  }

  QDialog::accept();
}

void DataTransformDialog::reject()
{
  QDialog::reject();
}




const QMap<QString,QVariant>& DataTransformDialog::readOptions() const
{
  return options_;
}
