/*******************************************************************************
**  FILE:
**  DESCRIPTION:
**
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/


#include "editorview2d.h"
#include "editor.h"
#include "pointmappergl.h"
#include "meshviewgl.h"

EditorView2D::EditorView2D(Editor *parent,
                           Point3D<float> hAxis, Point3D<float> vAxis)
  : EditorView      (parent)
  , hAxis_          ( hAxis )
  , vAxis_          ( vAxis )
{
  viewGL_->installEventFilter((QObject*)(this));

  Point3D<float> crossPoint = cross(vAxis_,hAxis_);
//  rotateAroundX(90*crossPoint.cy());
//  rotateAroundY(90*crossPoint.cx());

  viewGL_->setCameraUp(vAxis_);
  viewGL_->setCameraPos(crossPoint.mask()*(pointMapper_.worldSize()/2.0));


  Point3D<float> vMask = vAxis_.mask();
  Point3D<float> hMask = hAxis_.mask();

  connect(scrollH_,SIGNAL(valueChanged(int)),this,SLOT(scrollViewCameraH(int)));
  connect(scrollV_,SIGNAL(valueChanged(int)),this,SLOT(scrollViewCameraV(int)));

  connect(&pointMapper_,SIGNAL(sceneDepthsChanged(Point3D<float>))
          ,this ,SLOT(sceneDepthsChanged(Point3D<float>)));
  connect(&pointMapper_, SIGNAL(zoomFactorChanged(float)),
          viewGL_, SLOT(zoomFactorChanged()));

}




const Point3D<float> EditorView2D::viewMask()const
{
  return hAxis_.mask()+vAxis_.mask();
}


const Point3D<float>
EditorView2D::hAxis() const
{
  return hAxis_;
}


const Point3D<float>
EditorView2D::vAxis() const
{
  return vAxis_;
}



bool
EditorView2D::isHAxisInverted() const
{
  return (dot(hAxis_,Point3D<float>(1,1,1))<0);
}



bool
EditorView2D::isVAxisInverted() const
{
  return (dot(vAxis_,Point3D<float>(1,1,1))>0);
}




void
EditorView2D::setupGLView(QSize viewSize)
{
  int halfWorld = pointMapper_.mapGLToPixel(0);
  float orthoWidth = pointMapper_.mapPixelToGL(halfWorld+viewSize.width());
  float orthoHeight = pointMapper_.mapPixelToGL(halfWorld+viewSize.height());
  glOrtho(-orthoWidth/2.0, orthoWidth/2.0, -orthoHeight/2.0, orthoHeight/2.0,
          pointMapper_.drawingResolution(),
          pointMapper_.worldSize() + pointMapper_.drawingResolution());
//    glOrtho(0, orthoWidth, 0, orthoHeight,
//            pointMapper_.drawingResolution(), 100);
}


bool
EditorView2D::viewHasMouseTracking() const
{
  return viewGL_->hasMouseTracking();
}


void
EditorView2D::setViewMouseTracking(bool enable)
{
  viewGL_->setMouseTracking(enable);
}



//-----------------------------------------------------------------------------
//  PROTECTED METHODS
//-----------------------------------------------------------------------------


void
EditorView2D::wheelEvent(QWheelEvent *wheelEvent)
{
  if(wheelEvent->modifiers().testFlag(Qt::ControlModifier)){
    float zoom = pointMapper_.zoomFactor();

    zoom += (wheelEvent->delta()/120)*0.1;

    zoom = (zoom<0.1)? 0.1 : zoom;
    zoom = (zoom>5)? 5 : zoom;
//    pointMapper_.setZoomFactor(zoom);
  }
  else{
    if(wheelEvent->modifiers().testFlag(Qt::AltModifier)){
      QCoreApplication::sendEvent( static_cast<QObject*>(scrollH_),
                                   static_cast<QEvent*>(wheelEvent));
    }
    else{
      QCoreApplication::sendEvent( static_cast<QObject*>(scrollV_),
                                   static_cast<QEvent*>(wheelEvent));
    }
  }
}



bool EditorView2D::eventFilter(QObject *object, QEvent *event)
{
  // Mouse Move
  if(event->type() == QEvent::MouseMove){
    QMouseEvent* mouseEvent = static_cast<QMouseEvent*>(event);
    QPoint mousePos = mouseEvent->pos();
    pointMapper_.mouseTracker.setMovePosition(
    pointMapper_.mapFromViewToGL(mousePos,this),viewMask());
    emit mouseMoved(mousePos,mouseEvent->buttons(),mouseEvent->modifiers());
    return true;
  }
  // Mouse Press
  else if(event->type() == QEvent::MouseButtonPress){
    QMouseEvent* mouseEvent = static_cast<QMouseEvent*>(event);
    QPoint mousePos = mouseEvent->pos();
    Point3D<float> mouseGL = pointMapper_.mapFromViewToGL(mousePos,this);
    pointMapper_.mouseTracker.setPressPosition(mouseGL ,viewMask());
    pointMapper_.mouseTracker.setMovePosition(mouseGL ,viewMask());
    emit mousePressed(mousePos,mouseEvent->buttons(),mouseEvent->modifiers());
    return true;
  }
  // Mouse Release
  else if(event->type() == QEvent::MouseButtonRelease){
    QMouseEvent* mouseEvent = static_cast<QMouseEvent*>(event);
    QPoint mousePos = mouseEvent->pos();
    pointMapper_.mouseTracker.setReleasePosition(
    pointMapper_.mapFromViewToGL(mousePos,this),viewMask());
    emit mouseReleased(mousePos,mouseEvent->buttons(),mouseEvent->modifiers());
    return true;
  }
  return false;
}


//-----------------------------------------------------------------------------
//  PUBLIC SLOTS
//-----------------------------------------------------------------------------


void
EditorView2D::sceneDepthsChanged(Point3D<float> depths)
{
  float hDepthValue = dot(hAxis_.mask(),depths);
  float vDepthValue = dot(vAxis_.mask(),depths);

  Point3D<float> mask = hAxis_.mask() + vAxis_.mask();


  viewGL_->blockGLUpdate(true);
  viewGL_->setCameraPos(depths*mask,mask);
  viewGL_->setCameraAt(depths*mask,mask);
  viewGL_->blockGLUpdate(false);


  int hScrollVal = pointMapper_.mapGLToPixel(hDepthValue);
  int vScrollVal = pointMapper_.mapGLToPixel(vDepthValue);


  if(isVAxisInverted()){
    vScrollVal = pointMapper_.mapGLToPixel(pointMapper_.worldSize()/2) - vScrollVal;
  }

  if(isHAxisInverted()){
    hScrollVal = pointMapper_.mapGLToPixel(pointMapper_.worldSize()/2) - hScrollVal;
  }

  scrollH_->blockSignals(true);
  setHScrollValue(hScrollVal);
  scrollH_->blockSignals(false);

  scrollV_->blockSignals(true);
  setVScrollValue(vScrollVal);
  scrollV_->blockSignals(false);

  viewGL_->updateGL();
}


//-----------------------------------------------------------------------------
//  PRIVATE SLOTS
//-----------------------------------------------------------------------------



void
EditorView2D::scrollViewCameraH(int scrollValue)
{
  // If the the right of the H axis of this window represents + values
  int value;
  if(!isHAxisInverted()){
    value = scrollValue;
  }
  else{
    value = pointMapper_.mapGLToPixel(pointMapper_.worldSize()/2.0)-scrollValue;
  }

  emit hScrollChanged(scrollValue);

  Point3D<float> newDepths = pointMapper_.sceneDepths();
  newDepths = newDepths*(Point3D<float>(1,1,1)-hAxis().mask());
  newDepths += hAxis().mask()*pointMapper_.mapPixelToGL(value,false);
  pointMapper_.setSceneDepths(newDepths);
}



void
EditorView2D::scrollViewCameraV(int scrollValue)
{
  // If the the right of the H axis of this window represents + values
  int value;
  if(isVAxisInverted()){
    value = pointMapper_.mapGLToPixel(pointMapper_.worldSize()/2.0)-scrollValue;
  }
  else{
    value = scrollValue;
  }
  emit vScrollChanged(scrollValue);

  Point3D<float> newDepths = pointMapper_.sceneDepths();
  newDepths = newDepths*(Point3D<float>(1,1,1)-vAxis().mask());
  newDepths += vAxis().mask()*pointMapper_.mapPixelToGL(value,false);
  pointMapper_.setSceneDepths(newDepths);


}
