/*******************************************************************************
**  FILE:
**  DESCRIPTION:
**
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include "meshviewgl.h"
#include "pointmappergl.h"
#include "editorview.h"
#include "editorview2d.h"

//-----------------------------------------------------------------------------
// CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------


MeshViewGL::MeshViewGL(PointMapperGL& pointMapper, EditorView *editorView) :
    GLDisplayWidget(editorView),
    pointMapper_    (pointMapper),
    editorView_     (editorView),
    cameraPos_      (Point3D<GLfloat>(0.0,0.0,0.0)),
    cameraAt_       (Point3D<GLfloat>(0.0,0.0,0.0)),
    cameraUp_       (Point3D<GLfloat>(0.0,1.0,0.0)),
    cameraRight_    (Point3D<GLfloat>(1.0,0.0,0.0)),
    rotation_       (Point3D<GLfloat>(0.0,0.0,0.0))
{
  initializeLighting();
  connect(editorView_,SIGNAL(renderMeshItems(Mesh*)),
          this,SLOT(renderMesh(Mesh*)));

  connect(&(pointMapper_.mouseTracker),
          SIGNAL(pressPositionChanged(Point3D<float>&,Point3D<float>&)),
          this,
          SLOT(pressPointsChanged()));

}



MeshViewGL::~MeshViewGL()
{
}


//-----------------------------------------------------------------------------
//  PUBLIC METHODS
//-----------------------------------------------------------------------------


void
MeshViewGL::setCameraDistance(const GLfloat distance)
{
  cameraDistance_ = distance;
  updateGL();
}



void
MeshViewGL::setCameraPos(const Point3D<GLfloat> cameraPos,
                         const Point3D<float> changeVector)
{
  cameraPos_.rx() = (changeVector.cx() == 1)? cameraPos.cx() : cameraPos_.cx();
  cameraPos_.ry() = (changeVector.cy() == 1)? cameraPos.cy() : cameraPos_.cy();
  cameraPos_.rz() = (changeVector.cz() == 1)? cameraPos.cz() : cameraPos_.cz();
  updateGL();
}



void
MeshViewGL::moveCamera(const Point3D<GLfloat> deltaPos)
{
  cameraPos_ += deltaPos;
  updateGL();
}



void
MeshViewGL::setCameraAt(const Point3D<GLfloat> cameraAt,
                        const Point3D<float> changeVector)
{
  cameraAt_.rx() = (changeVector.cx() == 1)? cameraAt.cx() : cameraAt_.cx();
  cameraAt_.ry() = (changeVector.cy() == 1)? cameraAt.cy() : cameraAt_.cy();
  cameraAt_.rz() = (changeVector.cz() == 1)? cameraAt.cz() : cameraAt_.cz();
  updateGL();
}



void
MeshViewGL::setCameraUp(const Point3D<GLfloat> cameraUp,
                        const Point3D<float> changeVector)
{
  cameraUp_.rx() = (changeVector.cx() == 1)? cameraUp.cx() : cameraUp_.cx();
  cameraUp_.ry() = (changeVector.cy() == 1)? cameraUp.cy() : cameraUp_.cy();
  cameraUp_.rz() = (changeVector.cz() == 1)? cameraUp.cz() : cameraUp_.cz();
  updateGL();
}



void
MeshViewGL::setRotation(const Point3D<GLfloat> newRotation)
{
  rotation_ = newRotation;
  updateGL();
}



void
MeshViewGL::rotateAroundX(GLfloat angle)
{
  rotation_.rx() = angle;
  updateGL();
}



void
MeshViewGL::rotateAroundY(GLfloat angle)
{
  rotation_.ry() = angle;
  updateGL();
}



void
MeshViewGL::rotateAroundZ(GLfloat angle)
{
  rotation_.rz() = angle;
  updateGL();
}



GLfloat
MeshViewGL::getCameraDistance() const
{
  return cameraDistance_;
}



Point3D<GLfloat>
MeshViewGL::getCameraPos() const
{
  return cameraPos_;
}



Point3D<GLfloat>
MeshViewGL::getCameraAt() const
{
  return cameraAt_;
}



Point3D<GLfloat>
MeshViewGL::getCameraUp() const
{
  return cameraUp_;
}



Point3D<GLfloat>
MeshViewGL::getRotation() const
{
  return rotation_;
}



PointMapperGL& MeshViewGL::getPointMapper() const
{
  return pointMapper_;
}



EditorView* MeshViewGL::getEditorView() const
{
  return editorView_;
}



MeshItem*
MeshViewGL::pickItem(const QPoint& pos)
{
  MeshItem* pickedItem = 0;
  GLint itemPicked = pick(pos);
  if(itemPicked >= 0)
    pickedItem = meshItems_.at(itemPicked);

  return pickedItem;
}




void MeshViewGL::initGL()
{
  GLDisplayWidget::initGL();
  glClearColor(1.0, 1.0, 1.0, 0.0);
}



void
MeshViewGL::setupViewingVolume(QSize viewSize) const
{
  editorView_->setupGLView(viewSize);
}



void
MeshViewGL::setupGLLightingModel()
{
    glEnable(GL_LIGHTING);
    GLfloat ambientLight[] = { 0.3f, 0.3f, 0.3f, 1.0f };
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambientLight);

    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE | GL_SPECULAR);

    float specReflection[] = { 0.1f, 0.1f, 0.1f, 1.0f };
    glMaterialfv(GL_FRONT, GL_SPECULAR, specReflection);
    glMateriali(GL_FRONT, GL_SHININESS, 96);

  //    float black[4]={.05,.05,.05,0.05};
  //    float white[4]={0.1,0.1,0.1,0.1};
  //    glMaterialfv(GL_FRONT,GL_AMBIENT,black);
  //    glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,white);
}



void
MeshViewGL::draw() const
{
  bool gridVisible_ = false;
  // Todo: this portion should be handled using inheritance by subclassing
  // MeshViewGL for both the 2D and 3D case.
  EditorView2D* editor2D = qobject_cast<EditorView2D*>(editorView_);
  if(editor2D != 0){
    gridVisible_  = pointMapper_.gridVisible();
  }

  // TRANSFORMS
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();



  gluLookAt(cameraPos_.cx(),cameraPos_.cy(),cameraPos_.cz(),
            cameraAt_.cx(),cameraAt_.cy(),cameraAt_.cz(),
            cameraUp_.cx(),cameraUp_.cy(),cameraUp_.cz());

  // Adjust light position so that the items are consistantly lit
  QList<GLLight*> glLights = getLights();
  if(glLights.count() > 0){
    foreach(GLLight* light, glLights){
      glLightfv(light->glLight, GL_POSITION, light->pos);
    }
  }


  if(gridVisible_){
    drawGrid(editor2D);
  }


  //    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
  //    glLineWidth(3);
  if(meshItems_.count()>0){
    for(int i = 0;i<meshItems_.count();++i){
      glLoadName(i);
      meshItems_.at(i)->draw();
    }
  }

}



//-----------------------------------------------------------------------------
// PRIVATE METHODS
//-----------------------------------------------------------------------------



void
MeshViewGL::drawGrid(EditorView2D* editor) const
{  
  float drawingRes = pointMapper_.drawingResolution();
  float worldSize = pointMapper_.worldSize();
  float minorGrid = pointMapper_.minorGridSpacing()*drawingRes;
  float majorGrid = pointMapper_.majorGridSpacing()*minorGrid;

  int numMinorGridLines = worldSize/minorGrid;
  int numMajorGridLines = worldSize/majorGrid;
  int majorGridSpacing = pointMapper_.majorGridSpacing();

  QColor minorColor = Qt::lightGray;
  QColor majorColor = Qt::darkGray;


  float farWorldEdge = -worldSize/2.0;
  Point3D<GLfloat> glPoint, basePoint, conjVAxis, conjHAxis;
  basePoint = (Point3D<GLfloat>(1,1,1)-editor->viewMask())*(farWorldEdge);
  conjVAxis = (Point3D<GLfloat>(1,1,1) - editor->vAxis().mask()*2);
  conjHAxis = (Point3D<GLfloat>(1,1,1) - editor->hAxis().mask()*2);

  glDisable(GL_LIGHTING);
  glColor3f(minorColor.redF(),minorColor.greenF(),minorColor.blueF());
  glLineWidth(0.3);
  glBegin(GL_LINES);
  // Vertical Lines
  for(int i = 0; i < numMinorGridLines; ++i){
    glPoint = basePoint;
    if(i % majorGridSpacing == 0){
      glColor3f(majorColor.redF(),majorColor.greenF(),majorColor.blueF());
    }else{
      glColor3f(minorColor.redF(),minorColor.greenF(),minorColor.blueF());
    }
    glPoint += editor->hAxis().mask()*(farWorldEdge + i*minorGrid);
    glPoint += editor->vAxis()*(-farWorldEdge);
    glVertex3f(glPoint.cx(),glPoint.cy(),glPoint.cz());
    glPoint *= conjVAxis;
    glVertex3f(glPoint.cx(),glPoint.cy(),glPoint.cz());

    glPoint = basePoint;
    glPoint += editor->vAxis().mask()*(farWorldEdge + i*minorGrid);
    glPoint += editor->hAxis()*(farWorldEdge);
    glVertex3f(glPoint.cx(),glPoint.cy(),glPoint.cz());
    glPoint *= conjHAxis;
    glVertex3f(glPoint.cx(),glPoint.cy(),glPoint.cz());
  }
  glEnd();
  glEnable(GL_LIGHTING);

}


void
MeshViewGL::initializeLighting()
{

  GLLight* light = new GLLight;
  light->glLight = GL_LIGHT0;
  light->pos[0] = -5;  light->pos[1] = 5;
  light->pos[2] = 5;  light->pos[3] = 1;

  light->diffuse[0] = 0.15;  light->diffuse[1] = 0.15;
  light->diffuse[2] = 0.15;  light->diffuse[3] = 1;

  light->ambient[0] = 0.3;  light->ambient[1] = 0.3;
  light->ambient[2] = 0.3;  light->ambient[3] = 1;

  light->specular[0] = 1;  light->specular[1] = 1;
  light->specular[2] = 1;  light->specular[3] = 1;
  addLight(light);

//   Copy Last light and move to a different location
  light = new GLLight(*light);
  light->glLight = GL_LIGHT1;
  light->pos[0] = 10;  light->pos[1] = -5;
  light->pos[2] = 0;  light->pos[3] = 1;
  addLight(light);
}




//-----------------------------------------------------------------------------
//  SLOTS
//-----------------------------------------------------------------------------



void
MeshViewGL::renderMesh(Mesh *mesh)
{
  mesh->getMeshMutex().lockForRead();
      meshItems_ = mesh->items();
  mesh->getMeshMutex().unlock();
  updateGL();
}



void MeshViewGL::zoomFactorChanged()
{
  resizeViewPort(viewPortSize());
  updateGL();
}



void MeshViewGL::pressPointsChanged()
{
  //updateGL();
}

