/*******************************************************************************
**  FILE:         pointmappergl.cpp
**  DESCRIPTION:  PointMapperGL object implementation file.  See pointmappergl.h
**                for detailed object description.
**
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/
#include <QtCore>
#include "pointmappergl.h"
#include "editorview2D.h"

bool PointMapperGL::snapToGridFlag_ = true;

PointMapperGL::PointMapperGL(float worldSize,
                             float drawingResolution,
                             float zoomFactor)
  : worldSize_              (worldSize)
  , drawingResolution_      (drawingResolution)
  , zoomFactor_             (zoomFactor)
  , glSceneScale_           (1.0)
  , sceneDepths_            (Point3D<float>(worldSize_/2.0,
                                        worldSize_/2.0,
                                        worldSize_/2.0))
  , minorGridSpacing_       ( 10 )
  , majorGridSpacing_       ( 10 )
  , gridVisibleFlag_        ( true )
{
}



//-----------------------------------------------------------------------------
// PUBLIC METHODS
//-----------------------------------------------------------------------------


// These two functions only convert between pixels and the OpenGL world units
// they don't take into account the origin offset just unit conversion.

float PointMapperGL::convertPixelToGL(const int pixel,
                                      const bool snap) const
{
  float glCoord;
  if(snap == true){
    int snappedPixel = (pixel/minorGridSpacing_)*minorGridSpacing_;
    glCoord = snappedPixel*drawingResolution_/zoomFactor_;
  }else{
    glCoord = pixel*drawingResolution_/zoomFactor_;
  }
  return glCoord;
}


int PointMapperGL::convertGLToPixel(const float glCoord) const
{
  return qRound(glCoord/drawingResolution_*zoomFactor_);

}

// These functions take into account the origin offset of the world and actually
// map from the world to pixels and visa versa.

float
PointMapperGL::mapPixelToGL(const int pixel, const bool snap) const
{
  return convertPixelToGL(pixel,snap) - worldSize_/2.0;
}


int
PointMapperGL::mapGLToPixel(const float glCoord) const
{
  return convertGLToPixel(glCoord + worldSize_/2.0);
}




Point3D<float>
PointMapperGL::mapFromViewToGL(QPoint viewPoint, EditorView2D *editorView,
                               const bool snap) const
{

  QSize viewSize = editorView->viewPortSize();

  QPointF normalizedPoint = QPointF(viewPoint) - QPointF(viewSize.width()/2.0,
                                              viewSize.height()/2.0);
  normalizedPoint.ry()*= -1;

  Point3D<float> maskedPoint = editorView->vAxis()*
                               convertPixelToGL(normalizedPoint.y(),snap);
  maskedPoint += editorView->hAxis()*
                 convertPixelToGL(normalizedPoint.x(),snap);

  Point3D<float> retPoint = sceneDepths_ + maskedPoint;

  return retPoint;
}



QPoint
PointMapperGL::mapFromGLToView(const Point3D<float> glPoint,
                               EditorView2D *editorView) const
{
  QPoint retPoint;

  QSize viewSize = editorView->viewPortSize();

  float sceneDepthH = dot(sceneDepths_,editorView->hAxis().mask());
  float sceneDepthV = dot(sceneDepths_,editorView->vAxis().mask());

  float pointH = dot(glPoint,editorView->hAxis());
  float pointV = dot(glPoint,editorView->vAxis());

  retPoint.rx() = mapGLToPixel(sceneDepthH) - mapGLToPixel(pointH);
  retPoint.ry() = mapGLToPixel(pointV) - mapGLToPixel(sceneDepthV);

  retPoint = QPoint(viewSize.width()/2,viewSize.height()/2) - retPoint;

  return retPoint;
}



const float
PointMapperGL::worldSize() const
{
  return worldSize_;
}



const float
PointMapperGL::drawingResolution() const
{
  return drawingResolution_;
}



const float
PointMapperGL::zoomFactor() const
{
  return zoomFactor_;
}



const float
PointMapperGL::glSceneScale() const
{
  return glSceneScale_;
}



const Point3D<float>
PointMapperGL::sceneDepths() const
{
  return sceneDepths_;
}



const int
PointMapperGL::minorGridSpacing() const
{
  return minorGridSpacing_;
}



const int
PointMapperGL::majorGridSpacing() const
{
  return majorGridSpacing_;
}


const bool
PointMapperGL::gridVisible() const
{
  return gridVisibleFlag_;
}



const bool
PointMapperGL::snapToGrid() const
{
 return snapToGridFlag_;
}




//-----------------------------------------------------------------------------
//  PUBLIC SLOTS
//-----------------------------------------------------------------------------



void
PointMapperGL::setWorldSize(const float worldSize)
{
  worldSize_ = worldSize;
  emit worldSizeChanged( worldSize_ );
}



void
PointMapperGL::setDrawingResolution(const float drawingResolution)
{
  drawingResolution_ = drawingResolution;
  emit drawingResolutionChanged( drawingResolution_ );
}



void
PointMapperGL::setZoomFactor(float zoomFactor)
{
  zoomFactor_ = zoomFactor;
  emit zoomFactorChanged( zoomFactor_ );
}



void
PointMapperGL::setGLSceneScale(const float glSceneScale)
{
  glSceneScale_ = glSceneScale;
  emit glSceneScaleChanged( glSceneScale_ );
}



void
PointMapperGL::setSceneDepths(const Point3D<float> newDepths)
{
  blockSignals(true);
  setSceneDepthX(newDepths.cx());
  setSceneDepthY(newDepths.cy());
  setSceneDepthZ(newDepths.cz());
  blockSignals(false);
  emit sceneDepthXChanged( sceneDepths_.cx() );
  emit sceneDepthYChanged( sceneDepths_.cy() );
  emit sceneDepthZChanged( sceneDepths_.cz() );
  emit sceneDepthsChanged( sceneDepths_ );
}



void
PointMapperGL::setSceneDepthX(const float newDepth)
{
  sceneDepths_.rx() = newDepth;
  emit sceneDepthXChanged( sceneDepths_.cx() );
  emit sceneDepthsChanged( sceneDepths_ );
}



void
PointMapperGL::setSceneDepthY(const float newDepth)
{
  sceneDepths_.ry() = newDepth;
  emit sceneDepthYChanged( sceneDepths_.cy() );
  emit sceneDepthsChanged( sceneDepths_ );
}



void
PointMapperGL::setSceneDepthZ(const float newDepth)
{
  sceneDepths_.rz() = newDepth;
  emit sceneDepthZChanged( sceneDepths_.cz() );
  emit sceneDepthsChanged( sceneDepths_ );
}



void
PointMapperGL::setMinorGridSpacing(const int newSpacing)
{
  minorGridSpacing_ = newSpacing;
  emit minorGridSpacingChanged(minorGridSpacing_);
}



void
PointMapperGL::setMajorGridSpacing(const int newSpacing)
{
  majorGridSpacing_ = newSpacing;
  emit majorGridSpacingChanged(majorGridSpacing_);
}



void
PointMapperGL::setGridVisible(const bool visible)
{
  gridVisibleFlag_ = visible;
  emit gridVisibleChanged(gridVisibleFlag_);
}



void
PointMapperGL::setSnapToGrid(const bool snap)
{
  snapToGridFlag_ = snap;
  emit snapToGridChanged(snapToGridFlag_);
}
