/*******************************************************************************
**  FILE:
**  DESCRIPTION:
**
**
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include "meshsimulator.h"
#include "mesh.h"
#include "meshengine.h"
#include "meshsimulationthread.h"
#include "UI/Windows/mainwindow.h"


//-----------------------------------------------------------------------------
// CONSTRUCTORS
//-----------------------------------------------------------------------------


MeshSimulator::MeshSimulator(Mesh* mesh, QWidget *parent)
  : QWidget             (parent)
  , mesh_               (mesh)
  , simulationThread_   (new MeshSimulationThread(this))
  , meshEngine_ (0)
  , meshModifiedFlag_ (false)
  , statUpdateTimer_ (new QTimer(this))
{
  setAttribute(Qt::WA_DeleteOnClose);
  statUpdateTimer_->setInterval(500);
  createSimulationToolBar();
  createMeshEngineMenu();
  connect(this,SIGNAL(updateUI()),this,SLOT(uiUpdateRequested()));
  connect(mesh,SIGNAL(meshItemsModified(Mesh*)),this,SLOT(meshModified(Mesh*)));
  connect(simulationThread_,SIGNAL(simulationComplete()),
          this,SLOT(simulationComplete()));
  connect(statUpdateTimer_,SIGNAL(timeout()),this,SLOT(updateStats()));
  meshModified(mesh_);
  readSettings();
}



MeshSimulator::~MeshSimulator()
{
  writeSettings();
  simulationThread_->stopSimulation();
  delete simulationThread_;
}

//-----------------------------------------------------------------------------
// PUBLIC METHODS
//-----------------------------------------------------------------------------



QToolBar*
MeshSimulator::getSimulationToolBar() const
{
  return toolBarSimulation;
}



QMenu*
MeshSimulator::getMeshEngineMenu() const
{
  return meshEngineMenu;
}



MeshEngine*
MeshSimulator::getMeshEngine() const
{
  return meshEngine_;
}


//-----------------------------------------------------------------------------
// PRIVATE METHODS
//-----------------------------------------------------------------------------


void
MeshSimulator::readSettings()
{
  QSettings settings;
  changeTotalIterations(
        settings.value("meshSimulator/totalIterations", QVariant(2000)).toInt());
  changeUpdateInterval(
        settings.value("meshSimulator/updateInterval", QVariant(50)).toInt());
  changeDelayFactor(
        settings.value("meshSimulator/delayFactor", QVariant(50)).toInt());
}



void
MeshSimulator::writeSettings()
{
  QSettings settings;
  settings.setValue("meshSimulator/totalIterations",spinTotalIterations->value());
  settings.setValue("meshSimulator/updateInterval",spinUpdateInterval->value());
  settings.setValue("meshSimulator/delayFactor",spinDelayFactor->value());

}


void
MeshSimulator::createSimulationToolBar()
{
  toolBarSimulation     = new QToolBar(QString("Simulation Controls"));
  toolBarSimulation->setObjectName(QString("simControls"));

  buttonCreateMesh      = new QPushButton(tr("Create Mesh"));
  buttonRunSimulation   = new QPushButton(tr("Run"));
  buttonResetSimulation       = new QPushButton(tr("Reset"));
  buttonClearSimulation = new QPushButton(tr("Remove Mesh"));


  labelSimProgress      = new QLabel(tr("Simulation Progress "));
  labelTotalIterations  = new QLabel(tr("Total Iterations "));
  labelUpdateInterval   = new QLabel(tr("Update Interval "));
  labelDelayFactor      = new QLabel(tr("Delay Factor "));

  progressBarSimulation = new QProgressBar();
  progressBarSimulation->setOrientation(Qt::Horizontal);
  progressBarSimulation->setValue(0);

  sliderTotalIterations = new QSlider(Qt::Horizontal);
  sliderTotalIterations->setMinimum(1);
  sliderTotalIterations->setMaximum(50000);

  sliderUpdateInterval  = new QSlider(Qt::Horizontal);
  sliderUpdateInterval->setMinimum(1);
  sliderUpdateInterval->setMaximum(sliderTotalIterations->value());


  sliderDelayFactor     = new QSlider(Qt::Horizontal);
  sliderDelayFactor->setMinimum(0);
  sliderDelayFactor->setMaximum(2000);


  spinTotalIterations   = new QSpinBox();
  spinTotalIterations->setMinimum(sliderTotalIterations->minimum());
  spinTotalIterations->setMaximum(sliderTotalIterations->maximum());

  spinUpdateInterval    = new QSpinBox();
  spinUpdateInterval->setMinimum(sliderUpdateInterval->minimum());
  spinUpdateInterval->setMaximum(sliderUpdateInterval->maximum());

  spinDelayFactor       = new QSpinBox();
  spinDelayFactor->setMinimum(sliderDelayFactor->minimum());
  spinDelayFactor->setMaximum(sliderDelayFactor->maximum());


  // Add Widgets to toolbar

  toolBarSimulation->addWidget(buttonCreateMesh);
  toolBarSimulation->addWidget(buttonRunSimulation);
  toolBarSimulation->addWidget(buttonResetSimulation);
  toolBarSimulation->addWidget(buttonClearSimulation);

  toolBarSimulation->addSeparator();

  toolBarSimulation->addWidget(labelTotalIterations);
  toolBarSimulation->addWidget(sliderTotalIterations);
  toolBarSimulation->addWidget(spinTotalIterations);

  toolBarSimulation->addSeparator();

  toolBarSimulation->addWidget(labelUpdateInterval);
  toolBarSimulation->addWidget(sliderUpdateInterval);
  toolBarSimulation->addWidget(spinUpdateInterval);

  toolBarSimulation->addSeparator();

  toolBarSimulation->addWidget(labelDelayFactor);
  toolBarSimulation->addWidget(sliderDelayFactor);
  toolBarSimulation->addWidget(spinDelayFactor);

  toolBarSimulation->addSeparator();

  toolBarSimulation->addWidget(labelSimProgress);
  toolBarSimulation->addWidget(progressBarSimulation);


  connect(buttonCreateMesh,SIGNAL(clicked()),this,SLOT(createMesh()));
  connect(buttonRunSimulation,SIGNAL(clicked()),this,SLOT(startSimulation()));
  connect(buttonClearSimulation,SIGNAL(clicked()),this,SLOT(clearSimulation()));
  connect(buttonResetSimulation,SIGNAL(clicked()),this,SLOT(resetMesh()));

  connect(sliderTotalIterations,SIGNAL(valueChanged(int)),this,SLOT(changeTotalIterations(int)));
  connect(spinTotalIterations,SIGNAL(valueChanged(int)),this,SLOT(changeTotalIterations(int)));
  connect(sliderUpdateInterval,SIGNAL(valueChanged(int)),this,SLOT(changeUpdateInterval(int)));
  connect(spinUpdateInterval,SIGNAL(valueChanged(int)),this,SLOT(changeUpdateInterval(int)));
  connect(sliderDelayFactor,SIGNAL(valueChanged(int)),this,SLOT(changeDelayFactor(int)));
  connect(spinDelayFactor,SIGNAL(valueChanged(int)),this,SLOT(changeDelayFactor(int)));

}



void
MeshSimulator::createMeshEngineMenu()
{
  meshEngineMenu = new QMenu(tr("Mesh &Engines"));
  QList<MeshEngine*> engineList =  mesh_->getSupportedEngines();
  if(engineList.count()>0){
    QAction* actionPtr;
    QVariant actionData;
    foreach(MeshEngine* engine, engineList){
      actionPtr = meshEngineMenu->addAction(engine->getEngineName());
      // ToDo: The following should be changed so that it uses the QVariant
      // from value methods which require Q_DECLARE_METATYPE with the MeshEngine
      // Class.  Using void pointer for now.
      actionData.setValue((void*)engine);
      actionPtr->setData(actionData);
      actionPtr->setCheckable(true);
      actionPtr->setEnabled(engine->isCompatible());

      connect(actionPtr,SIGNAL(toggled(bool)),this,SLOT(meshEngineSelectionChanged()));
    }
//     Select the first compatible engine
    foreach(actionPtr, meshEngineMenu->actions()){
      if(actionPtr->isEnabled()){
        actionPtr->setChecked(true);
        break;
      }
    }
    meshEngineMenu->setEnabled(true);
  }
  else{
    meshEngineMenu->setEnabled(false);
  }
}



void
MeshSimulator::setMeshModifiedFlag(const bool flag)
{
  meshModifiedFlag_ = flag;
  emit updateUI();
}


void
MeshSimulator::setTotalIterationsEnabled(bool enabled)
{
  spinTotalIterations->setEnabled(enabled);
  sliderTotalIterations->setEnabled(enabled);
}


void
MeshSimulator::setUpdateIntervalEnabled(bool enabled)
{
  spinUpdateInterval->setEnabled(enabled);
  sliderUpdateInterval->setEnabled(enabled);
}


void
MeshSimulator::setDelayFactorEnabled(bool enabled)
{
  spinDelayFactor->setEnabled(enabled);
  sliderDelayFactor->setEnabled(enabled);
}

//-----------------------------------------------------------------------------
// PROTECTED METHODS
//-----------------------------------------------------------------------------


void
MeshSimulator::changeEvent(QEvent *event)
{
  if(event->type() == QEvent::EnabledChange){
    bool enabled_ = isEnabled();

    buttonCreateMesh->setEnabled(enabled_);
    buttonRunSimulation->setEnabled(enabled_);
    buttonClearSimulation->setEnabled(enabled_);

    labelTotalIterations->setEnabled(enabled_);
    sliderTotalIterations->setEnabled(enabled_);
    spinTotalIterations->setEnabled(enabled_);

    labelUpdateInterval->setEnabled(enabled_);
    sliderUpdateInterval->setEnabled(enabled_);
    spinUpdateInterval->setEnabled(enabled_);

    labelDelayFactor->setEnabled(enabled_);
    sliderDelayFactor->setEnabled(enabled_);
    spinDelayFactor->setEnabled(enabled_);

    labelSimProgress->setEnabled(enabled_);
    progressBarSimulation->setEnabled(enabled_);
  }
}



//-----------------------------------------------------------------------------
// PRIVATE SLOTS
//-----------------------------------------------------------------------------


void
MeshSimulator::meshEngineSelectionChanged()
{
  QAction* senderAction = qobject_cast<QAction*>(QObject::sender());
  MeshEngine* senderEngine = 0;
  if(senderAction != 0){
     senderEngine = (MeshEngine*)(senderAction->data().value<void*>());
  }
  setMeshEngine(senderEngine);
}


void
MeshSimulator::setMeshEngine(MeshEngine *meshEngine)
{
  if(meshEngine_ != meshEngine){
    if(meshEngine_ != 0){
      disconnect(meshEngine_,SIGNAL(voltageDataChanged()),
                 this,SLOT(meshEngineDataUpdated()));
    }
    MeshEngine* actionEngine;
    foreach(QAction* actionPtr, meshEngineMenu->actions()){
      actionEngine = (MeshEngine*)(actionPtr->data().value<void*>());
      actionPtr->blockSignals(true);
      actionPtr->setChecked(meshEngine == actionEngine);
      actionPtr->blockSignals(false);
    }
    if(meshEngine_ != 0){
      blockSignals(true);
      clearSimulation();
      blockSignals(false);
    }
    meshEngine_ = meshEngine;
    if(meshEngine_ != 0){
      connect(meshEngine_,SIGNAL(voltageDataChanged()),
              this, SLOT(meshEngineDataUpdated()));
    }
    emit meshEngineChanged(meshEngine_);
    emit updateUI();
  }
}



void
MeshSimulator::meshEngineDataUpdated()
{
  MeshEngine* senderEngine = qobject_cast<MeshEngine*>(QObject::sender());
  emit updateDisplays(senderEngine);
}



void
MeshSimulator::uiUpdateRequested()
{
  if(meshEngine_ != 0){
    if(meshModifiedFlag_){
      buttonCreateMesh->setEnabled(true);
    }else{
      buttonCreateMesh->setEnabled(false);
    }
    if(meshEngine_->meshSize() != MeshSize(0,0,0)){
      bool hasRan = meshEngine_->getCurrentIteration() > 0;
      buttonResetSimulation->setEnabled(hasRan);
      setTotalIterationsEnabled(!hasRan);
      setUpdateIntervalEnabled(!hasRan);
      setDelayFactorEnabled(!hasRan);
      buttonClearSimulation->setEnabled(true);
      if(simulationThread_->isRunning()){
        buttonRunSimulation->setText(tr("Pause"));
        buttonRunSimulation->disconnect(this,SLOT(startSimulation()));
        connect(buttonRunSimulation,SIGNAL(clicked()),this,SLOT(pauseSimulation()));
      }else{
        buttonRunSimulation->setText(tr("Run"));
        buttonRunSimulation->disconnect(this,SLOT(pauseSimulation()));
        connect(buttonRunSimulation,SIGNAL(clicked()),this,SLOT(startSimulation()));
      }
      buttonRunSimulation->setEnabled(true);
    }else{
      buttonClearSimulation->setEnabled(false);
      buttonRunSimulation->setEnabled(false);
      buttonResetSimulation->setEnabled(false);
    }
  }else{
    buttonCreateMesh->setEnabled(false);
    buttonClearSimulation->setEnabled(false);
    buttonRunSimulation->setEnabled(false);
    buttonResetSimulation->setEnabled(false);
  }
}



void
MeshSimulator::iterationFinished(int iteration)
{
//  if(iteration != totalIterations_){
//    progressPercent_ = qRound((float)iteration/totalIterations_*100);
//    setProgress(tr("Running..."),progressPercent_);
//  }else{
//    progressPercent_ = 100;
//    setProgress(tr("Complete..."),progressPercent_);
//  }
//  emit iterationChanged(iteration);
}


void
MeshSimulator::simulationComplete()
{

  statUpdateTimer_->stop();
  updateStats();
}


void MeshSimulator::updateStats()
{
  int iteration = meshEngine_->getCurrentIteration();
  if(iteration != totalIterations_){
    progressPercent_ = qRound((float)iteration/totalIterations_*100);
    setProgress(tr("Running..."),progressPercent_);
  }else{
    progressPercent_ = 100;
    setProgress(tr("Complete..."),progressPercent_);
  }
  emit iterationChanged(iteration);
}




//-----------------------------------------------------------------------------
// PUBLIC SLOTS
//-----------------------------------------------------------------------------


void
MeshSimulator::editorModeChanged(Editor::EditorMode mode)
{
  int dimensions = 0;
  if(mode == Editor::Edit2D){
    dimensions = 2;
  }
  else if(mode == Editor::Edit3D){
    dimensions = 3;
  }
  MeshEngine* engine;
  foreach(QAction* action, meshEngineMenu->actions())
  {
    engine = (MeshEngine*)(action->data().value<void*>());
    if(engine->dimensions() != dimensions){
      if(engine == meshEngine_){
        setMeshEngine(0);
      }
      action->setDisabled(true);
    }else{
      action->setEnabled(true);
    }
  }
}



void
MeshSimulator::startSimulation()
{
  buttonRunSimulation->setText(tr("Pause"));
  buttonRunSimulation->disconnect(this,SLOT(startSimulation()));
  connect(buttonRunSimulation,SIGNAL(clicked()),this,SLOT(pauseSimulation()));
  if(simulationThread_->isRunning()){
    simulationThread_->resumeSimulation();
  }else{
    qDebug()<<"Running new simulation run.";
    totalIterations_ = spinTotalIterations->value();
    simulationThread_->simulate(meshEngine_,
                                totalIterations_,
                                spinUpdateInterval->value(),
                                spinDelayFactor->value());
    buttonResetSimulation->setEnabled(true);
    setTotalIterationsEnabled(false);
    setUpdateIntervalEnabled(false);
    setDelayFactorEnabled(false);
  }
  statUpdateTimer_->start();
  setProgress(tr("Running"),progressPercent_);
}



void
MeshSimulator::pauseSimulation()
{
  statUpdateTimer_->stop();
  buttonRunSimulation->setText(tr("Run"));
  buttonRunSimulation->disconnect(this,SLOT(pauseSimulation()));
  connect(buttonRunSimulation,SIGNAL(clicked()),this,SLOT(startSimulation()));
  simulationThread_->pauseSimulation();
  setProgress(tr("Paused"),progressPercent_);
}



void
MeshSimulator::resetMesh()
{
  statUpdateTimer_->stop();
  simulationThread_->stopSimulation();
  meshEngine_->resetSimulationData();
  setMeshModifiedFlag(false);
  setProgress(tr("Ready"),0);
  emit updateUI();
}



void
MeshSimulator::clearSimulation()
{
  statUpdateTimer_->stop();
  simulationThread_->stopSimulation();
  meshEngine_->deleteSimulationData();
  setMeshModifiedFlag(true);
  setProgress(tr("No Mesh Created"),0);
  emit updateUI();
}



void
MeshSimulator::createMesh()
{
  setProgress(tr("Discretizing..."),0);
  meshEngine_->createSimulation();
  setMeshModifiedFlag(false);
  setProgress(tr("Ready"),0);

}



void
MeshSimulator::changeTotalIterations(int totalIterations)
{
  spinTotalIterations->blockSignals(true);
  sliderTotalIterations->blockSignals(true);

  spinTotalIterations->setValue(totalIterations);
  sliderTotalIterations->setValue(totalIterations);

  spinUpdateInterval->setMaximum(totalIterations);
  sliderUpdateInterval->setMaximum(totalIterations);

  spinTotalIterations->blockSignals(false);
  sliderTotalIterations->blockSignals(false);
}



void
MeshSimulator::changeUpdateInterval(int updateInterval)
{
  spinUpdateInterval->blockSignals(true);
  sliderUpdateInterval->blockSignals(true);

  spinUpdateInterval->setValue(updateInterval);
  sliderUpdateInterval->setValue(updateInterval);

  spinUpdateInterval->blockSignals(false);
  sliderUpdateInterval->blockSignals(false);
}



void
MeshSimulator::changeDelayFactor(int delayFactor)
{
  spinDelayFactor->blockSignals(true);
  sliderDelayFactor->blockSignals(true);

  spinDelayFactor->setValue(delayFactor);
  sliderDelayFactor->setValue(delayFactor);

  spinDelayFactor->blockSignals(false);
  sliderDelayFactor->blockSignals(false);
}



void
MeshSimulator::setProgress(QString text, int value)
{
  progressPercent_ = value;
  labelSimProgress->setText(text);
  progressBarSimulation->setValue(value);
}



void
MeshSimulator::meshModified(Mesh *mesh)
{
  if(mesh->items().count() > 0){
    setMeshModifiedFlag(true);
  }else{
    setMeshModifiedFlag(false);
  }
}
