/*******************************************************************************
**  FILE:
**  DESCRIPTION:
**
**
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include <QtGui>
#include "mainwindow.h"
// Managers
#include "UI/MeshEditors/editor.h"
#include "UI/MeshSimulator/meshsimulator.h"
#include "UI/Animator/animator.h"
#include "UI/DataViewers/probeviewer.h"
// QtOclMesh for computations
#include "qtoclmesh.h"



//-----------------------------------------------------------------------------
// CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------


MainWindow::MainWindow(QWidget *parent)
  : QMainWindow(parent)
  , meshEditor     (0)
  , editorWindow   (0)
{
    // The following line sets the attribute for the main window
    // so that it will call it's destructor when it is closed.
    // However with the current implmenetation of QSharedPointers
    // to all objects, some of the objects should be pointed to using
    // QExplicitelySharedDataPointer objects because they are actually
    // sharing data and it throws seg faults when the destructor is called.
    //setAttribute(Qt::WA_DeleteOnClose);

    // SET THE CORE APPLICATION SETTINGS WHICH ARE USED WHEN SAVING APPLICATION SETTINGS
    QCoreApplication::setOrganizationName("University of Victoria");
    QCoreApplication::setOrganizationDomain("uvic.ca");
    QCoreApplication::setApplicationName("ElectroSim");


    mdiArea = new QMdiArea;
    mdiArea->setViewMode(QMdiArea::TabbedView);
    mdiArea->setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
    mdiArea->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);
    setCentralWidget(mdiArea);
    connect(mdiArea, SIGNAL(subWindowActivated(QMdiSubWindow*)),
	    this, SLOT(mdiWindowActivated(QMdiSubWindow*)));

    windowMapper = new QSignalMapper(this);
    connect(windowMapper, SIGNAL(mapped(QWidget*)),
	    this, SLOT(setActiveSubWindow(QWidget*)));


    // Creates objects that need to be present for connecting signals
    preInitialize();
    // Connect the signals of the modules together
    connectSignals();
    // Read in the saved application settings
    readSettings();
    // Initialize the application
    initialize();
    createMenuBar();



}



MainWindow::~MainWindow()
{

    cursor.clear();
    delete meshSimulator;
    delete meshEditor;
}



//-----------------------------------------------------------------------------
// PROTECTED METHODS
//-----------------------------------------------------------------------------


void
MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
//        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}



void
MainWindow::closeEvent(QCloseEvent *event){
  mdiArea->closeAllSubWindows();
  if(mdiArea->currentSubWindow()){
    event->ignore();
  }else{
    writeSettings();
    meshSimulator->close();
    event->accept();
  }
}



//-----------------------------------------------------------------------------
// PRIVATE METHODS
//-----------------------------------------------------------------------------


void
MainWindow::preInitialize(){
    // Create the Mesh object
    mesh = new DielectricMesh(this);


    // SIMULATION CONTROLS
    meshSimulator = new MeshSimulator(mesh,this);
    addToolBar(meshSimulator->getSimulationToolBar());

//    // ANIMATOR
//    meshAnimator = new Animator(this);
//    animatorWindow =  mdiArea->addSubWindow(meshAnimator,
//                                          meshAnimator->windowFlags());
//    animatorWindow->setMaximumHeight(meshAnimator->maximumHeight());
//    animatorWindow->setMaximumWidth(meshAnimator->maximumWidth());
//    animatorWindow->setWindowTitle(meshAnimator->windowTitle());
//    animatorWindow->setWindowState(meshAnimator->windowState());
//    connect(meshSimulator,SIGNAL(updateDisplays(MeshEngine*)),
//            meshAnimator,SLOT(updateAnimation(MeshEngine*)));


    // EDITOR
    meshEditor = new Editor(mesh,this);
    connect(meshEditor,SIGNAL(setCursorPos(QPoint)),this,SLOT(setCursorPos(QPoint)));
    editorWindow =  mdiArea->addSubWindow(meshEditor, meshEditor->windowFlags());
    editorWindow->setWindowTitle(meshEditor->windowTitle());
    editorWindow->setWindowState(Qt::WindowMaximized);

    // EDITOR TOOLBAR
    addDockWidget(Qt::LeftDockWidgetArea,
                  meshEditor->getMeshItemDockWidget(),
                  Qt::Vertical);
    meshEditor->getMeshItemDockWidget()->show();


    // ADD STATUS BAR
    statusBar = new QStatusBar(this);
    statusBar->setObjectName(QString::fromUtf8("statusbar"));
    setStatusBar(statusBar);


}



void
MainWindow::connectSignals()
{
  connect(meshEditor,SIGNAL(editorModeChanged(Editor::EditorMode)),
          meshSimulator,SLOT(editorModeChanged(Editor::EditorMode)));

  connect(meshSimulator,SIGNAL(iterationChanged(int)),
          this,SLOT(iterationChanged(int)));

}



void
MainWindow::initialize(){
    // Create cursor object and attach it to the central widget
    cursor = QSharedPointer<QCursor>(new QCursor());
    setCursor(*cursor);

    meshEditor->setEditorMode(Editor::Edit2D);



    // Set the scene of the editor view to the MeshScene created above
//    ui->editorGraphicsView->setScene(meshScene.data());
    //ui->editorGraphicsView->setDragMode(QGraphicsView::RubberBandDrag);
    //ui->editorGraphicsView->setRubberBandSelectionMode(Qt::IntersectsItemShape);

//    ui->tabView->setCurrentWidget(ui->Editor);

//    undoStack = itemManager->getUndoStackPointer();
//    // Create the undo / redo actions in the edit menu
//    ui->menuEdit->removeAction(ui->actionUndo);
//    ui->menuEdit->removeAction(ui->actionRedo);
//    ui->actionUndo = undoStack->createUndoAction(this,tr("&Undo"));
//    ui->actionUndo->setShortcuts(QKeySequence::Undo);
//    ui->actionUndo->setIcon(QIcon(":Images/Images/undo.png"));

//    ui->actionRedo = undoStack->createRedoAction(this,tr("&Redo"));
//    ui->actionRedo->setShortcuts(QKeySequence::Redo);
//    ui->actionRedo->setIcon(QIcon(":Images/Images/redo.png"));

//    ui->menuEdit->insertAction(ui->actionDeleteItem,ui->actionRedo);
//    ui->menuEdit->insertAction(ui->actionRedo,ui->actionUndo);
//    ui->menuEdit->insertSeparator(ui->actionDeleteItem);


//    undoView = QSharedPointer<QUndoView>(new QUndoView(undoStack.data(),ui->commandDockContents));
//    undoView->setMaximumSize(50000,50000);
//    ui->commandDockContentLayout->addWidget(undoView.data());

}



void
MainWindow::readSettings()
{
    QSettings settings;
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("state").toByteArray(),0);
}



void
MainWindow::writeSettings()
{
    QSettings settings;
    settings.setValue("geometry",saveGeometry());
    settings.setValue("state",saveState(0));
}



void
MainWindow::createFileMenu()
{
  menuFile = new QMenu(tr("&File"),this);
  QAction* action;
  QIcon icon;

  icon = QIcon();
  icon.addFile(QString::fromUtf8(":/Images/Images/newNormalOn.png"), QSize(), QIcon::Normal, QIcon::On);
  icon.addFile(QString::fromUtf8(":/Images/Images/newNormalOff.png"), QSize(), QIcon::Normal, QIcon::Off);
  action = new QAction(icon,tr("&New File"),this);
  action->setShortcut(QKeySequence::New);
  action->setObjectName(QString("actionNewFile"));
  connect(action,SIGNAL(triggered()),this,SLOT(fileNew()));
  menuFile->addAction(action);
  action->setEnabled(false);


  icon = QIcon();
  icon.addFile(QString::fromUtf8(":/Images/Images/openNormalOn.png"), QSize(), QIcon::Normal, QIcon::On);
  icon.addFile(QString::fromUtf8(":/Images/Images/openNormalOff.png"), QSize(), QIcon::Normal, QIcon::Off);
    action = new QAction(icon,tr("&Open File"),this);
  action->setShortcut(QKeySequence::Open);
  action->setObjectName(QString("actionOpenFile"));
  connect(action,SIGNAL(triggered()),this,SLOT(fileOpen()));
  menuFile->addAction(action);

  icon = QIcon();
  icon.addFile(QString::fromUtf8(":/Images/Images/saveNormalOn.png"), QSize(), QIcon::Normal, QIcon::On);
     action = new QAction(icon,tr("&Save"),this);
  action->setShortcut(QKeySequence::Save);
  action->setObjectName(QString("actionSaveFile"));
  connect(action,SIGNAL(triggered()),this,SLOT(fileSave()));
  menuFile->addAction(action);
  action->setEnabled(false);

  action = new QAction(tr("Save &As..."),this);
  action->setShortcut(QKeySequence::SaveAs);
  action->setObjectName(QString("actionSaveFileAs"));
  connect(action,SIGNAL(triggered()),this,SLOT(fileSaveAs()));
  menuFile->addAction(action);

  connect(menuFile,SIGNAL(aboutToShow()),this,SLOT(beforeShowFileMenu()));
}



void
MainWindow::createWindowMenu()
{
  menuWindow = new QMenu(tr("&Window"),this);
  menuOpenWindows = new QMenu(tr("&Open Windows"),menuWindow);

  QAction* action;

  action = new QAction(tr("New Probe &Viewer"),this);
  action->setObjectName(QString("actionNewProbeViewer"));
  connect(action,SIGNAL(triggered()),this,SLOT(createNewProbeViewer()));
  menuWindow->addAction(action);

  action = new QAction(tr("New &Animator"),this);
  action->setObjectName(QString("actionNewAnimator"));
  connect(action,SIGNAL(triggered()),this,SLOT(createNewAnimator()));
  menuWindow->addAction(action);

  menuWindow->addSeparator();

  menuWindow->addMenu(menuOpenWindows);

  connect(menuWindow,SIGNAL(aboutToShow()),this,SLOT(beforeShowWindowMenu()));

}



void
MainWindow::createMenuBar()
{
  createFileMenu();
  createWindowMenu();
  menuBar = new QMenuBar();
  setMenuBar(menuBar);
  // Todo:  add file menu
  menuBar->addMenu(menuFile);
  menuBar->addMenu(meshEditor->getItemActionMenu());
  menuBar->addMenu(meshEditor->getEditorMenu());
  QMenu* newMenu = menuBar->addMenu(tr("&Simulation"));
  newMenu->addMenu(meshSimulator->getMeshEngineMenu());
  menuBar->addMenu(menuWindow);

}



void
MainWindow::updateWindowMenu()
{

}



//-----------------------------------------------------------------------------
// PRIVATE SLOTS
//-----------------------------------------------------------------------------

void
MainWindow::beforeShowFileMenu()
{
  qDebug()<<"about to show file menu.";
}


void
MainWindow::beforeShowWindowMenu()
{
  menuOpenWindows->clear();
  QAction* action;
  foreach(QMdiSubWindow* subWindow, mdiArea->subWindowList()){
    action = new QAction(subWindow->windowTitle(),menuOpenWindows);
    connect(action,SIGNAL(triggered()),subWindow,SLOT(showMaximized()));
    menuOpenWindows->addAction(action);
  }
}


void
MainWindow::createNewProbeViewer()
{
  // This should be changed so that it doesn't name two probe viewers with the
  // same name if a probe viewer has been closed.
  int idNum = 0;
  ProbeViewer* probeViewerTest;
  foreach(QMdiSubWindow* subWindow, mdiArea->subWindowList()){
    probeViewerTest = qobject_cast<ProbeViewer*>(subWindow->widget());
    if(probeViewerTest != 0){
      ++idNum;
    }
  }
  QString windowTitle = tr("Probe Viewer");

  if(idNum > 0){
    windowTitle.append(tr(" %1").arg(idNum));
  }

  ProbeViewer* probeViewer = new ProbeViewer(mesh,this);
  QMdiSubWindow* subWindow = mdiArea->addSubWindow(probeViewer);
  subWindow->setWindowTitle(windowTitle);
  subWindow->setWindowState(Qt::WindowActive);
  connect(meshSimulator,SIGNAL(updateDisplays(MeshEngine*)),probeViewer,SLOT(updateGraph()));
  probeViewer->updateGraph();
}



void
MainWindow::createNewAnimator()
{

  // This should be changed so that it doesn't name two animators with the
  // same name if an animator has been closed.
  int idNum = 0;
  Animator* animatorTest;
  foreach(QMdiSubWindow* subWindow, mdiArea->subWindowList()){
    animatorTest = qobject_cast<Animator*>(subWindow->widget());
    if(animatorTest != 0){
      ++idNum;
    }
  }
  QString windowTitle = tr("Animator");

  if(idNum > 0){
    windowTitle.append(tr(" %1").arg(idNum));
  }

  Animator* meshAnimator = new Animator(this);
  QMdiSubWindow* animatorWindow =  mdiArea->addSubWindow(meshAnimator);
  animatorWindow->setWindowTitle(windowTitle);
  animatorWindow->setWindowState(Qt::WindowActive);

  connect(meshSimulator,SIGNAL(updateDisplays(MeshEngine*)),
          meshAnimator,SLOT(updateAnimation(MeshEngine*)));
  if(meshSimulator->getMeshEngine()!=0){
    meshAnimator->updateAnimation(meshSimulator->getMeshEngine());
  }
}



//-----------------------------------------------------------------------------
// PUBLIC SLOTS
//-----------------------------------------------------------------------------


void
MainWindow::mdiWindowActivated(QMdiSubWindow *subWindow)
{
  if(subWindow==editorWindow){
    meshEditor->getMeshItemDockWidget()->setEnabled(true);
  }else{
    meshEditor->getMeshItemDockWidget()->setEnabled(false);
  }
}



void
MainWindow::setCursorShape(Qt::CursorShape shape)
{
    cursor->setShape(shape);
    setCursor(*cursor);
}



void
MainWindow::setCursorPos(QPoint pos)
{
    cursor->setPos(pos);
}



void
MainWindow::setStatus(const QString status, const int timeOut)
{
  statusBar->showMessage(status,timeOut);
}



void
MainWindow::iterationChanged(int iteration)
{
  setStatus(tr("Iteration %1").arg(iteration));
}


//---------------   File slots	----------------//



void
MainWindow::fileNew()
{
    QMessageBox::information(this,QString("New File"),
                             QString("New File code will go here."),QMessageBox::Ok);
}



void
MainWindow::fileOpen()
{
  QFileDialog::Options options;
  QString selectedFilter = tr("ElectroSim Files (*.esf)");
  QString directory = QDesktopServices::storageLocation(QDesktopServices::DesktopLocation);
  QString fileName = QFileDialog::getOpenFileName(this,
                                                  tr("Open Mesh ..."),
                                                  directory,
                                                  tr("ElectroSim Files (*.esf)"),
                                                  &selectedFilter,
                                                  options);
  if (!fileName.isEmpty()){
    QFile dataFile(fileName);
    if(!dataFile.open(QIODevice::ReadOnly)){
      // ToDo: show dialog box informing user
      qDebug()<<"Could not open File.";
    }else{
      QDataStream in(&dataFile);
      in.setVersion(QDataStream::Qt_4_7);
      in >> mesh;
//      int itemCount;
//      in >> itemCount;
//      MeshItem* inputItem;
//      qDebug()<<"Item Count:"<<itemCount;
//      for(int i = 0; i < itemCount; ++i){
//        in >> *inputItem;
//        qDebug()<<"Class Name:"<<inputItem->metaObject()->className();
//      }
      dataFile.close();
    }
  }}



void
MainWindow::fileSave()
{
    QMessageBox::information(this,QString("Save File"),
                             QString("Save File code will go here."),QMessageBox::Ok);
}



void
MainWindow::fileSaveAs()
{
  QFileDialog::Options options;
  QString selectedFilter = tr("ElectroSim Files (*.esf)");
  QString directory = QDesktopServices::storageLocation(QDesktopServices::DesktopLocation);
  QString fileName = QFileDialog::getSaveFileName(this,
                                                  tr("Save Mesh to file..."),
                                                  directory.append("\\untitled.esf"),
                                                  tr("ElectroSim Files (*.esf)"),
                                                  &selectedFilter,
                                                  options);
  if (!fileName.isEmpty()){
    QFile dataFile(fileName);
    if(!dataFile.open(QIODevice::WriteOnly)){
      // ToDo: show dialog box informing user
      qDebug()<<"Could not open File.";
    }else{
      QDataStream out(&dataFile);
      out.setVersion(QDataStream::Qt_4_7);
//      QList<MeshItem*> itemsToSave = mesh->items();
//      int itemCount = itemsToSave.count();
//      out<<itemCount;
//      for(int i = 0; i < itemCount; ++i){
//        out<<*(itemsToSave.at(i));
//      }
      out << mesh;
      dataFile.close();
    }
  }
}




void
MainWindow::setActiveSubWindow(QWidget *window)
{
    if (!window)
        return;
    QMdiSubWindow* subWindow = qobject_cast<QMdiSubWindow *>(window);
    mdiArea->setActiveSubWindow(subWindow);
    window->setWindowState(Qt::WindowMaximized);
}
